/*
 *  TT's Bit Ops Plugin
 * ---------------------
 *
 * A REALbasic 2.0 plugin for performing Bit Operations
 * Works with 68K, PPC, Carbon and x86 (Windows)
 *
 * The accompanying project file was created using CodeWarrior Pro 4 (IDE 3.3)
 *
 * This code is freeware, you may use and extend it as you like.
 * It was written on Jan 19, 2000 by Thomas Tempelmann.
 *
 * More RB-related stuff can be found here: <http://www.tempel.org/rb/>
 *
 * Enjoy and contribute!
 */


#include "rb_plugin.h"


//---------------
// configuration
//---------------

#define UseRBStringsForBuffering 0
		// This flag is used for the new bitset class that is added by this plugin.
		// set to 0 if you want to use Mac OS heap memory, set to 1 for using RB memory.
		// Both have its odds and ends: Using heap memory will not crash if the buffer
		// required is larger than the available memory, but it may run out if you
		// create classes later in the process.
		// My recommendation: Create all the required Bitset classes at start of
		// your app, and avoid creating them later on the fly. In you can handle
		// that, use the Mac OS heap memory by setting this flag to 0.
		// And test you app well under low memory situations!
		// The best solution would be if these problems get properly addressed
		// in RB and its Plugin API, but until then you have to be careful.

//-----------------
// helper routines
//-----------------

typedef unsigned char* BFPtr;


static inline long Min (long a, long b)
{
	if (a <= b) return a; else return b;
}

static inline long Max (long a, long b)
{
	if (a >= b) return a; else return b;
}

static void copyFewBytes (unsigned char* s, unsigned char* d, unsigned long len)
// copy routine for short amount of data
{
	if (len) {
		if (d > s && d < (s+len)) {
			// need to copy backwards
			d += len;
			s += len;
			do { *--d = *--s; } while (--len);
		} else {
			do { *d++ = *s++; } while (--len);
		}
	}
}

static inline void copyBytes (unsigned char * s0, unsigned char* d, unsigned long len)
{
	#if TARGET_OS_MAC
		if (len < 100) {
			copyFewBytes (s0, d, len);
		} else {
			BlockMoveData (s0, d, len);
		}
	#else
		// For Windows, I do not know a fast copy routine for larger amounts
		// of data such as BlockMove for Mac OS.
		copyFewBytes (s0, d, len);
	#endif
}

static void clearBytes (unsigned char *d, unsigned long len)
{
	if (len) {
		do { *d++ = 0; } while (--len);
	}
}

// now comes some helper routines for allocation, disposal and access to the bitset buffers

#if UseRBStringsForBuffering || !(TARGET_OS_MAC)
	
	static inline BFPtr GetBuffer (void* p)
	{
		return (BFPtr) ((REALstring)p)->CString();
	}
	
	static void* AllocBuffer (BFPtr srcData, long size)
	{
		BFPtr newData;
		newData = (BFPtr) REALBuildString ((Ptr)srcData, size);
		if (newData && !srcData) {
			clearBytes (GetBuffer (newData), size);
		}
		return (void*) newData;
	}
	
	static void FreeBuffer (void* p)
	{
		REALUnlockString ((REALstring)p);
	}

#else
	
	static inline BFPtr GetBuffer (void* p)
	{
		return (BFPtr) p;
	}
	
	static void* AllocBuffer (BFPtr srcData, long size)
	{
		BFPtr newData;
		if (srcData) {
			newData = (BFPtr) NewPtr (size);
			if (newData) {
				copyBytes (srcData, newData, size);
			}
		} else {
			newData = (BFPtr) NewPtrClear (size);
		}
		return (void*) newData;
	}

	static void FreeBuffer (void* p)
	{
		DisposePtr ((Ptr)p);
	}

#endif

//------------------------
// Methods added globally
//------------------------


	//------------------
	// Integer routines
	//------------------

static long shiftA (long v, long sh)
// signed shift
{
	if (sh < 0) {
		if (sh < -31) return (v < 0) ? -1 : 0;	// Aaron Ballman notified me (TT) that this check is necessary because CPUs would take the "sh" value modulo 32, and C isn't preventing this, either
		return v >> -sh;
	} else {
		if (sh > 31) return 0;	// Aaron Ballman notified me (TT) that this check is necessary because CPUs would take the "sh" value modulo 32, and C isn't preventing this, either
		return v << sh;
	}
}

static unsigned long shiftL (unsigned long v, long sh)
// unsigned shift
{
	if (sh < 0) {
		if (sh < -31) return 0;	// Aaron Ballman notified me (TT) that this check is necessary because CPUs would take the "sh" value modulo 32, and C isn't preventing this, either
		return v >> -sh;
	} else {
		if (sh > 31) return 0;	// Aaron Ballman notified me (TT) that this check is necessary because CPUs would take the "sh" value modulo 32, and C isn't preventing this, either
		return v << sh;
	}
}

static unsigned long bitNot (unsigned long v)
{
	return ~v;
}

static unsigned long bitDiff (unsigned long v1, unsigned long v2)
{
	return v1 & ~v2;
}

static unsigned long bitNOr (unsigned long v1, unsigned long v2)
{
	return ~v1 | ~v2;
}

static unsigned long bitNAnd (unsigned long v1, unsigned long v2)
{
	return ~v1 & ~v2;
}

static unsigned long bitMask[33] =
{
	0x00000000,
	0x00000001, 0x00000003, 0x00000007, 0x0000000F, 0x0000001F, 0x0000003F, 0x0000007F, 0x000000FF,
	0x000001FF, 0x000003FF, 0x000007FF, 0x00000FFF, 0x00001FFF, 0x00003FFF, 0x00007FFF, 0x0000FFFF,
	0x0001FFFF, 0x0003FFFF, 0x0007FFFF, 0x000FFFFF, 0x001FFFFF, 0x003FFFFF, 0x007FFFFF, 0x00FFFFFF,
	0x01FFFFFF, 0x03FFFFFF, 0x07FFFFFF, 0x0FFFFFFF, 0x1FFFFFFF, 0x3FFFFFFF, 0x7FFFFFFF, 0xFFFFFFFF
};

static unsigned long rotate (unsigned long v, long count, long offset, long width)
{
	if (count == 0) return v;
	if (offset < 0) { width += offset; offset = 0; }
	if ((width+offset) > 32) width = 32-offset;
	if (width <= 1) return v;
	unsigned long rotMask = bitMask[width] << offset;
	long c;
	if (count < 0) {
		// example: 100001 ROT -2 = 100001 ROTR 2 = 001000 | 010000 = 011000
		c = -count % width;
	} else {
		// example: 100001 ROT +2 = 100001 ROTR 4 = 000010 | 000100 = 000110
		c = (width-count) % width;
	}
	return (v & ~rotMask) | ((((v & rotMask) >> c) | ((v & (bitMask[c] << offset)) << (width-c))) & rotMask);
}

static unsigned long bitVal (unsigned long n)
{
	if (n > 31) return 0;
	return 1L << n;
}

static unsigned long bitIncl (unsigned long v, unsigned long n)
{
	if (n > 31) return v;
	return v | (1L << n);
}

static unsigned long bitExcl (unsigned long v, unsigned long n)
{
	if (n > 31) return v;
	return v & ~(1L << n);
}

static long bitIn (unsigned long v, unsigned long n)
{
	if (n > 31) return false;
	return (v & (1L << n)) != 0;
}

	//--------------------
	// Bit field routines
	//--------------------

static unsigned char bfMask[9] =
{
	0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF
};

static void bfClr1 (BFPtr addr, long bitsInBitfield)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*addr++ = 0;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		*addr = *addr & ~bfMask[rem];
	}
}

static void bfClr2 (BFPtr addr, long ofs, long numBitsToClear)
{
	addr += ofs >> 3;
	ofs &= 7;
	if (ofs) {
		if (numBitsToClear <= 0) return;
		long bytes = (numBitsToClear+ofs-1) >> 3;
		unsigned char mask = bfMask[ofs];
		short rem = (numBitsToClear+ofs) & 7;
		if (bytes == 0) {
			if (rem) {
				*addr = *addr & (mask | ~bfMask[rem]);
			} else {
				*addr = *addr & mask;
			}
		} else {
			*addr = *addr & mask;
			++addr;
			while (--bytes > 0) {
				*addr++ = 0;
			}
			if (rem) {
				*addr = *addr & ~bfMask[rem];
			}
		}
	} else {
		bfClr1 (addr, numBitsToClear);
	}
}

static long bfEmpty (BFPtr addr, long bitsInBitfield)
{
	if (bitsInBitfield <= 0) return 1; // return true
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		if (*addr++) return 0; // return false
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		return (*addr & bfMask[rem]) == 0;
	}
	return 1; // return true
}

static void bfNot (BFPtr addr, long bitsInBitfield)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*addr = ~*addr;
		++addr;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		*addr = *addr ^ bfMask[rem];
	}
}

static void bfAnd (BFPtr src1, long bitsInBitfield, BFPtr src2, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = *src1++ & *src2++;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | (*src1 & *src2 & mask);
	}
}

static void bfOr (BFPtr src1, long bitsInBitfield, BFPtr src2, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = *src1++ | *src2++;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | ((*src1 | *src2) & mask);
	}
}

static void bfXOr (BFPtr src1, long bitsInBitfield, BFPtr src2, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = *src1++ ^ *src2++;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | ((*src1 ^ *src2) & mask);
	}
}

static void bfNOr (BFPtr src1, long bitsInBitfield, BFPtr src2, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = ~*src1++ | ~*src2++;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | ((~*src1 | ~*src2) & mask);
	}
}

static void bfNAnd (BFPtr src1, long bitsInBitfield, BFPtr src2, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = ~*src1++ & ~*src2++;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | ((~*src1 & ~*src2) & mask);
	}
}

static void bfDiff (BFPtr src1, long bitsInBitfield, BFPtr src2, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = *src1++ & ~*src2++;	// this clears the bitsInBitfield that are inluded in src2
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | ((*src1 & ~*src2) & mask);
	}
}

static void bfCpy (BFPtr src, long bitsInBitfield, BFPtr dest)
{
	if (bitsInBitfield <= 0) return;
	long bytes = bitsInBitfield >> 3;
	while (--bytes >= 0) {
		*dest++ = *src++;
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		*dest = (*dest & ~mask) | (*src & mask);
	}
}

static long bitCmp (unsigned char a, unsigned char b)
{
	for (int i = 0; i < 8; ++i) {
		if ((a & 1) != (b & 1)) return i;
		a >>= 1;
		b >>= 1;
	}
	// oops, where's the difference?!
	return -1;
}

static long bfCmp (BFPtr src, long bitsInBitfield, BFPtr dest)
// returns the bit number where the two fields differ, or -1 if there's no difference
{
	if (bitsInBitfield <= 0) return -1;
	long bytes = bitsInBitfield >> 3;
	BFPtr start = dest;
	while (--bytes >= 0) {
		if (*dest++ != *src++) {
			return bitCmp (*--dest, *--src) | ((dest - start) << 3);	// found a diff
		}
	}
	long rem = bitsInBitfield & 7;
	if (rem) {
		unsigned char mask = bfMask[rem];
		if ((*dest & mask) != (*src & mask)) {
			return bitCmp (*dest, *src) | ((dest - start) << 3);	// found a diff
		}
	}
	return -1;	// no diff
}

static long bitFind (short b, short start, short max, short findSetBit)
{
	for (short i = start; i < max; ++i) {
		if ((b >> i) & 1 == findSetBit) {
			return i;
		}
	}
	return -1;
}

static long bfFind (BFPtr src, long bitsInBitfield, long start, unsigned long findSetBit)
// searches the bit field, starting at "start" bit number, for a bit that
// is set or clear. returns its position, or -1 if none is found.
{
	BFPtr startPtr = src;
	long remBits = bitsInBitfield;
	if (start >= remBits) return -1;
	if (start) {
		if (start < 0) start = 0;
		src += start >> 3;
		remBits -= start & ~7;
		start &= 7;
	}

	if (findSetBit > 1) findSetBit = 1;

	long bytes = remBits >> 3;
	short rem = remBits & 7;
	if (start) {
		// we must start a a non-0 pos in the first byte
		short max;
		if (!bytes) {
			// this is also the last byte
			max = rem;
		} else {
			max = 8;
		}
		short i = bitFind (*src++, start, max, findSetBit);
		if (i >= 0) {
			return ((src-startPtr) << 3) + i;
		}
		--bytes;
	}

	// now check the remaining bytes but the last one
	if (findSetBit) {
		while (--bytes >= 0) {
			if (*src++) {
				// found a byte with a set bit
				return ((src-startPtr) << 3) | bitFind (*--src, 0, 8, 1);
			}
		}
	} else {
		while (--bytes >= 0) {
			if (*src++ != 0xFF) {
				// found a byte with a set clear
				return ((src-startPtr) << 3) | bitFind (*--src, 0, 8, findSetBit);
			}
		}
	}

	if (rem) {
		return ((src-startPtr) << 3) | bitFind (*--src, 0, rem, 0);
	}

	return -1;	// no diff found
}

static long bfGet (BFPtr src, long bitsInBitfield, long bitNum)
{
	if (bitsInBitfield <= 0 || bitNum < 0 || bitNum >= bitsInBitfield) return 0;
	BFPtr bp = &src[bitNum>>3];
	return (*bp & (1 << (bitNum & 7))) != 0;
}

static void bfSet (BFPtr src, long bitsInBitfield, long bitNum, long value)
{
	if (bitsInBitfield <= 0 || bitNum < 0 || bitNum >= bitsInBitfield) return;
	BFPtr bp = &src[bitNum>>3];
	if (value != 0) {
		*bp |= 1 << (bitNum & 7);
	} else {
		*bp &= ~(1 << (bitNum & 7));
	}
}

static void bfCpyShft (BFPtr srcPtr, long srcOfs, BFPtr destPtr, long destOfs, long bitsToCopy)
// This is the super duper Bit Copy routine: It can copy any amount of bits from any
// bit location to a different bit location.
// Don't ask me how this routine works. I had a hard time to get it working,
// with considering all the possible cases, and I hope I never have to work
// on it again.
{
	if (bitsToCopy <= 0) return;
	
	srcPtr += srcOfs >> 3;
	srcOfs &= 7;
	destPtr += destOfs >> 3;
	destOfs &= 7;

	short shiftAmount = destOfs - srcOfs;
	
	long srcBytes = ((bitsToCopy + srcOfs - 1) >> 3);

	Boolean overlap = (destPtr > srcPtr && destPtr < (srcPtr+((bitsToCopy+7)>>3))); // if true, backwards copy is necessary

	unsigned char lastByte, lastMask, firstByte, firstMask;
	short n;

	if (shiftAmount == 0) {
		// no need to shift inside bytes - we just move whole bytes

		if (srcPtr == destPtr) return;
		firstMask = bfMask[destOfs];
		n = (srcOfs+bitsToCopy) & 7;
		if (n==0) n = 8;
		lastMask = ~bfMask[n];
		if (srcBytes == 0) {
			// only one byte
			lastMask |= firstMask;
			*destPtr = (*destPtr & lastMask) | (*srcPtr & ~lastMask);
		} else {
			// copy first and last bytes with their masks
			if (overlap) {
				destPtr[srcBytes] = (destPtr[srcBytes] & lastMask) | (srcPtr[srcBytes] & ~lastMask);
				if (srcBytes>1) {
					copyBytes (srcPtr+1, destPtr+1, srcBytes-1);
				}
				destPtr[0] = (destPtr[0] & firstMask) | (srcPtr[0] & ~firstMask);
			} else {
				destPtr[0] = (destPtr[0] & firstMask) | (srcPtr[0] & ~firstMask);
				if (srcBytes>1) {
					copyBytes (srcPtr+1, destPtr+1, srcBytes-1);
				}
				destPtr[srcBytes] = (destPtr[srcBytes] & lastMask) | (srcPtr[srcBytes] & ~lastMask);
			}
		}

	} else {
		// bytes need individual shifting.
		
		long destBytes = ((bitsToCopy + destOfs - 1) >> 3);
		Boolean sameByteCnt = (srcBytes == destBytes);

		unsigned short b2;
	
		if (shiftAmount > 0) {
			// left shift (that means: towards higher bit positions)
			//assert (srcBytes <= destBytes);
			firstByte = destPtr[0];
			firstMask = bfMask[destOfs];
			lastByte = destPtr[destBytes];
			n = (srcOfs+bitsToCopy+shiftAmount) & 7;
			if (n==0) n = 8;
			lastMask = ~bfMask[n];
			if (destBytes == 0) {
				// only one byte is affected
				b2 = *srcPtr++;
				lastMask |= firstMask;
				if (srcBytes) {
					b2 |= *srcPtr << 8;
				}
				*destPtr = (lastByte & lastMask) | ((b2 << shiftAmount) & ~lastMask);
			} else {
				//assert (srcBytes <= destBytes);
				if (overlap) {
					short shiftRev = 8 - shiftAmount;
					srcPtr += srcBytes;
					destPtr += destBytes;
					b2 = *srcPtr;
					if (sameByteCnt) {
						b2 = b2 << 8 | *--srcPtr;
					}
					*destPtr = (lastByte & lastMask) | ((b2 >> shiftRev) & ~lastMask);
					while (--destBytes > 0) {
						b2 = b2 << 8 | *--srcPtr;
						*--destPtr = b2 >> shiftRev;
					}
					*--destPtr = (firstByte & firstMask) | ((b2 << shiftAmount) & ~firstMask);
				} else {
					b2 = *srcPtr++ << shiftAmount;
					*destPtr++ = (firstByte & firstMask) | (b2 & ~firstMask);
					while (--destBytes > 0) {
						b2 = *srcPtr++ << shiftAmount | b2 >> 8;
						*destPtr++ = b2;
					}
					b2 >>= 8;
					if (sameByteCnt) {
						b2 |= *srcPtr++ << shiftAmount;
					}
					*destPtr = (lastByte & lastMask) | (b2 & ~lastMask);
				}
			}
		} else {
			// right shift (that means: towards lower bit positions)
			//assert (srcBytes >= destBytes);
			shiftAmount = -shiftAmount;
			firstByte = destPtr[0];
			firstMask = bfMask[destOfs];
			lastByte = destPtr[destBytes];
			n = (srcOfs-shiftAmount+bitsToCopy) & 7;
			if (n==0) n = 8;
			lastMask = ~bfMask[n];
			if (destBytes == 0) {
				// only one byte is affected
				b2 = *srcPtr++;
				lastMask |= firstMask;
				if (srcBytes) {
					b2 |= *srcPtr << 8;
				}
				*destPtr = (lastByte & lastMask) | ((b2 >> shiftAmount) & ~lastMask);
			} else {
				if (overlap) {
					srcPtr += srcBytes;
					destPtr += destBytes;
					b2 = *srcPtr;
					if (!sameByteCnt) {
						b2 = b2 << 8 | *--srcPtr;
					}
					*destPtr = (lastByte & lastMask) | ((b2 >> shiftAmount) & ~lastMask);
					b2 = b2 << 8 | *--srcPtr;
					while (--destBytes > 0) {
						*--destPtr = b2 >> shiftAmount;
						b2 = b2 << 8 | *--srcPtr;
					}
					*--destPtr = (firstByte & firstMask) | ((b2 >> shiftAmount) & ~firstMask);
				} else {
					b2 = *srcPtr++;
					b2 |= *srcPtr++ << 8;
					*destPtr++ = (firstByte & firstMask) | ((b2 >> shiftAmount) & ~firstMask);
					while (--destBytes > 0) {
						b2 = *srcPtr++ << 8 | b2 >> 8;
						*destPtr++ = b2 >> shiftAmount;
					}
					b2 >>= 8;
					if (!sameByteCnt) {
						b2 |= *srcPtr++ << 8;
					}
					*destPtr = (lastByte & lastMask) | ((b2 >> shiftAmount) & ~lastMask);
				}
			}
		}
	}
}

static void bfShift (BFPtr src, long bitsInBitfield, long count)
{
	if (bitsInBitfield <= 0 || count == 0) return;
	if (count > 0) {
		long n = bitsInBitfield-count;
		if (n <= 0) return;
		bfCpyShft (src, 0, src, count, n);
		bfClr1 (src, count);
	} else {
		long n = bitsInBitfield+count;
		if (n <= 0) return;
		bfCpyShft (src, -count, src, 0, n);
		bfClr2 (src, n, -count);
	}
}

static REALmethodDefinition pluginMethodArray[] =
{
		// Integer (32 bit) operations

	{ (REALproc) shiftA,  REALnoImplementation, "ArithmeticShift(value as Integer, count as Integer) as Integer" },
	{ (REALproc) shiftL,  REALnoImplementation, "LogicalShift(value as Integer, count as Integer) as Integer" },
	{ (REALproc) rotate,  REALnoImplementation, "BitwiseRotate(value as Integer, count as Integer, offset as Integer, width as Integer) as Integer" },
	{ (REALproc) bitNot,  REALnoImplementation, "BitwiseNot(value as Integer) as Integer" },
	{ (REALproc) bitVal,  REALnoImplementation, "BitVal(bitNumber as Integer) as Integer" },
	{ (REALproc) bitIncl, REALnoImplementation, "BitIncl(value as Integer, bitNumber as Integer) as Integer" },
	{ (REALproc) bitExcl, REALnoImplementation, "BitExcl(value as Integer, bitNumber as Integer) as Integer" },
	{ (REALproc) bitIn,   REALnoImplementation, "BitIsSet(value as Integer, bitNumber as Integer) as Boolean" },
	{ (REALproc) bitDiff, REALnoImplementation, "BitwiseDiff(val1 as Integer, val2 as Integer) as Integer" },
	{ (REALproc) bitNOr,  REALnoImplementation, "BitwiseNOr(val1 as Integer, val2 as Integer) as Integer" },
	{ (REALproc) bitNAnd, REALnoImplementation, "BitwiseNAnd(val1 as Integer, val2 as Integer) as Integer" },

		// Generic operations on any bit array in memory

	{ (REALproc) bfGet,    REALnoImplementation, "BitfieldBit(address as Integer, sizeOfBitfield as Integer, bitNum as Integer) as Boolean" },
	{ (REALproc) bfSet,    REALnoImplementation, "BitfieldSet(address as Integer, sizeOfBitfield as Integer, bitNum as Integer, val as Boolean)" },
	{ (REALproc) bfClr1,   REALnoImplementation, "BitfieldClear(address as Integer, sizeOfBitfield as Integer)" },
	{ (REALproc) bfEmpty,  REALnoImplementation, "BitfieldEmpty(address as Integer, sizeOfBitfield as Integer) as Boolean" },
	{ (REALproc) bfNot,    REALnoImplementation, "BitfieldNot(address as Integer, sizeOfBitfield as Integer)" },
	{ (REALproc) bfAnd,    REALnoImplementation, "BitfieldAnd(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer, destAddr as Integer)" },
	{ (REALproc) bfOr,     REALnoImplementation, "BitfieldOr(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer, destAddr as Integer)" },
	{ (REALproc) bfXOr,    REALnoImplementation, "BitfieldXOr(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer, destAddr as Integer)" },
	{ (REALproc) bfNOr,    REALnoImplementation, "BitfieldNOr(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer, destAddr as Integer)" },
	{ (REALproc) bfNAnd,   REALnoImplementation, "BitfieldNAnd(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer, destAddr as Integer)" },
	{ (REALproc) bfCpy,    REALnoImplementation, "BitfieldCopy(srcAddr as Integer, sizeOfBitfield as Integer, destAddr as Integer)" },
	{ (REALproc) bfDiff,   REALnoImplementation, "BitfieldDiff(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer, destAddr as Integer)" },
	{ (REALproc) bfShift,  REALnoImplementation, "BitfieldShift(address as Integer, sizeOfBitfield as Integer, shiftCount as Integer)" },
	{ (REALproc) bfFind,   REALnoImplementation, "BitfieldFind(address as Integer, sizeOfBitfield as Integer, startBit as Integer, value as Boolean) as Integer" },
	{ (REALproc) bfCmp,    REALnoImplementation, "BitfieldCompare(srcAddr1 as Integer, sizeOfBitfield as Integer, srcAddr2 as Integer) as Integer" },
	
	{ (REALproc) bfCpyShft,REALnoImplementation, "BitfieldCopyBits(srcAddr as Integer, srcBitOfs as Integer, destAddr as Integer, destBitOfs as Integer, bitsToCopy as Integer)" },
	{ (REALproc) bfClr2,   REALnoImplementation, "BitfieldClearBits(address as Integer, bitOfs as Integer, bits as Integer)" },
};


//------------------
// New Bitset class
//------------------


#define bsName "TTsBitSet" // the name of this new class

extern REALclassDefinition bitsetClass;	// forward declaration

struct bitsetClassData
{
	long		bits;
	long		bytes;
	void*		data;	// I use a RB String as a container for the Bitset data
};

static Boolean bsConstructor2 (REALobject instance, long bits, BFPtr srcData)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	
	if (bits <= 0) {
		return false;
	}

	me->bits = bits;
	me->bytes = (me->bits + 7) >> 3;
	me->data = AllocBuffer (srcData, me->bytes);
	if (!me->data) {
		// If we get here, the data buffer could not be allocated.
		me->bits = me->bytes = 0;
		return false;
	}
	return true;
}

static void bsConstructor (REALobject instance, long bits)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	Boolean ok = bsConstructor2 (instance, bits, (unsigned char*)NULL);
	if (!ok) {
		// unfortunately, this does not work in order to have NIL returned
		// from the "new" call (as of RB v2.1a25):
		//   REALUnlockObject (instance);
	}
}

static void bitsetClassDestructor (REALobject instance)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);

	if (me->data) {
		me->bits = me->bytes = 0;
		FreeBuffer (me->data);
		me->data = nil;
	}
}


static long bsBitsGet (REALobject instance, long param)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	return me->bits;
}

static BFPtr bsAddrGet (REALobject instance, long param)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	return GetBuffer (me->data);
}

static long bsSizeGet (REALobject instance, long param)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	return me->bytes;
}

static long bsBitGet (REALobject instance, long bitNum)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	return bfGet (GetBuffer(me->data), me->bits, bitNum);
}

static void bsBitSet (REALobject instance, long bitNum, long val)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	bfSet (GetBuffer(me->data), me->bits, bitNum, val);
}

static REALobject bsClone (REALobject instance)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	REALobject theClone = REALnewInstance (bsName);
	if (theClone) {
 		Boolean ok = bsConstructor2 (theClone, me->bits, GetBuffer(me->data));
 		if (!ok) {
			REALUnlockObject (theClone);
 			theClone = nil;
 		}
	}
	return theClone;
}

static long bsEmpty (REALobject instance)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	return bfEmpty (GetBuffer(me->data), me->bits);
}

static void bsClear (REALobject instance)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	bfClr1 (GetBuffer(me->data), me->bits);
}

static void bsNot (REALobject instance)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	bfNot (GetBuffer(me->data), me->bits);
}

static long bsEqual (REALobject inst1, REALobject inst2)
{
	ClassData (bitsetClass, inst1, bitsetClassData, me);
	ClassData (bitsetClass, inst2, bitsetClassData, him);
	if (me->bits != him->bits) {
		return 0;	// false
	}
	return bfCmp (GetBuffer(me->data), me->bits, GetBuffer(him->data)) == -1;
}

static void bsCopy (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfCpy (GetBuffer(src->data), bits, GetBuffer(me->data));
}

static void bsOr1 (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfOr (GetBuffer(src->data), bits, GetBuffer(me->data), GetBuffer(me->data));
}

static void bsOr2 (REALobject instance, REALobject srcInst1, REALobject srcInst2)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst1, bitsetClassData, src1);
	ClassData (bitsetClass, srcInst2, bitsetClassData, src2);
	long bits = Min (me->bits, Min (src1->bits, src2->bits));
	bfOr (GetBuffer(src1->data), bits, GetBuffer(src2->data), GetBuffer(me->data));
}

static void bsAnd1 (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfAnd (GetBuffer(src->data), bits, GetBuffer(me->data), GetBuffer(me->data));
}

static void bsAnd2 (REALobject instance, REALobject srcInst1, REALobject srcInst2)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst1, bitsetClassData, src1);
	ClassData (bitsetClass, srcInst2, bitsetClassData, src2);
	long bits = Min (me->bits, Min (src1->bits, src2->bits));
	bfAnd (GetBuffer(src1->data), bits, GetBuffer(src2->data), GetBuffer(me->data));
}

static void bsNOr1 (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfNOr (GetBuffer(src->data), bits, GetBuffer(me->data), GetBuffer(me->data));
}

static void bsNOr2 (REALobject instance, REALobject srcInst1, REALobject srcInst2)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst1, bitsetClassData, src1);
	ClassData (bitsetClass, srcInst2, bitsetClassData, src2);
	long bits = Min (me->bits, Min (src1->bits, src2->bits));
	bfNOr (GetBuffer(src1->data), bits, GetBuffer(src2->data), GetBuffer(me->data));
}

static void bsNAnd1 (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfNAnd (GetBuffer(src->data), bits, GetBuffer(me->data), GetBuffer(me->data));
}

static void bsNAnd2 (REALobject instance, REALobject srcInst1, REALobject srcInst2)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst1, bitsetClassData, src1);
	ClassData (bitsetClass, srcInst2, bitsetClassData, src2);
	long bits = Min (me->bits, Min (src1->bits, src2->bits));
	bfNAnd (GetBuffer(src1->data), bits, GetBuffer(src2->data), GetBuffer(me->data));
}

static void bsXOr1 (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfXOr (GetBuffer(src->data), bits, GetBuffer(me->data), GetBuffer(me->data));
}

static void bsXOr2 (REALobject instance, REALobject srcInst1, REALobject srcInst2)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst1, bitsetClassData, src1);
	ClassData (bitsetClass, srcInst2, bitsetClassData, src2);
	long bits = Min (me->bits, Min (src1->bits, src2->bits));
	bfXOr (GetBuffer(src1->data), bits, GetBuffer(src2->data), GetBuffer(me->data));
}

static void bsDiff1 (REALobject instance, REALobject srcInst)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long bits = Min (me->bits, src->bits);
	bfDiff (GetBuffer(src->data), bits, GetBuffer(me->data), GetBuffer(me->data));
}

static void bsDiff2 (REALobject instance, REALobject srcInst1, REALobject srcInst2)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	ClassData (bitsetClass, srcInst1, bitsetClassData, src1);
	ClassData (bitsetClass, srcInst2, bitsetClassData, src2);
	long bits = Min (me->bits, Min (src1->bits, src2->bits));
	bfDiff (GetBuffer(src1->data), bits, GetBuffer(src2->data), GetBuffer(me->data));
}

static void bsShift (REALobject instance, long count)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	bfShift (GetBuffer(me->data), me->bits, count);
}

static void bsBtClr (REALobject instance, long startBit, long count)
{
	ClassData (bitsetClass, instance, bitsetClassData, me);
	if (startBit < 0) {
		count -= startBit;
		startBit = 0;
	}
	long rem = me->bits - startBit;
	if (rem <= 0) return;
	if (count > rem) count = rem;
	bfClr2 (GetBuffer(me->data), startBit, count);
}

static void bsBtCp2 (REALobject destInst, REALobject srcInst, long srcBitOfs, long destBitOfs, long numBits)
{
	ClassData (bitsetClass, destInst, bitsetClassData, me);
	ClassData (bitsetClass, srcInst, bitsetClassData, src);
	long rem, maxBits;

	maxBits = Min (me->bits, src->bits);
	
	// validate all parms (offsets must not be negative)
	if (numBits <= 0) return;
	if (srcBitOfs < destBitOfs) {
		if (srcBitOfs < 0) {
			numBits -= srcBitOfs;
			destBitOfs -= srcBitOfs;
			srcBitOfs = 0;
		}
		rem = numBits - destBitOfs;
	} else {
		if (destBitOfs < 0) {
			numBits -= destBitOfs;
			srcBitOfs -= destBitOfs;
			destBitOfs = 0;
		}
		rem = numBits - srcBitOfs;
	}
	if (rem <= 0) return;
	if (numBits > rem) numBits = rem;

	bfCpyShft (GetBuffer(src->data), srcBitOfs, GetBuffer(me->data), destBitOfs, numBits);
}

static void bsBtCp1 (REALobject instance, long srcBitOfs, long destBitOfs, long numBits)
{
	bsBtCp2 (instance, instance, srcBitOfs, destBitOfs, numBits);
}

static REALproperty bitsetClassProperties[] = {
	{ nil, "Bits", "Integer", 0, (REALproc) bsBitsGet, REALnoImplementation },
	{ nil, "Size", "Integer", 0, (REALproc) bsSizeGet, REALnoImplementation },
	{ nil, "Address", "Integer", 0, (REALproc) bsAddrGet, REALnoImplementation },
};

static REALmethodDefinition bitsetClassMethods[] = {
	{ (REALproc) bsConstructor, REALnoImplementation, bsName"(bits as Integer)"},
	{ (REALproc) bsBitGet, (REALproc) bsBitSet, "Bit(bitNum as Integer) as Boolean"},
	{ (REALproc) bsClone, REALnoImplementation, "Clone() as "bsName},
	{ (REALproc) bsEmpty, REALnoImplementation, "Empty() as Boolean"},
	{ (REALproc) bsEqual, REALnoImplementation, "Equals(other as "bsName") as Boolean"},
	{ (REALproc) bsClear, REALnoImplementation, "Clear()"},
	{ (REALproc) bsCopy,  REALnoImplementation, "Copy(source as "bsName")"},
	{ (REALproc) bsNot,   REALnoImplementation, "Invert()"},
	{ (REALproc) bsOr1,   REALnoImplementation, "BitsOr(source as "bsName")"},
	{ (REALproc) bsOr2,   REALnoImplementation, "BitsOr(src1 as "bsName", src2 as "bsName")"},
	{ (REALproc) bsAnd1,  REALnoImplementation, "BitsAnd(source as "bsName")"},
	{ (REALproc) bsAnd2,  REALnoImplementation, "BitsAnd(src1 as "bsName", src2 as "bsName")"},
	{ (REALproc) bsNOr1,  REALnoImplementation, "BitsNOr(source as "bsName")"},
	{ (REALproc) bsNOr2,  REALnoImplementation, "BitsNOr(src1 as "bsName", src2 as "bsName")"},
	{ (REALproc) bsNAnd1, REALnoImplementation, "BitsNAnd(source as "bsName")"},
	{ (REALproc) bsNAnd2, REALnoImplementation, "BitsNAnd(src1 as "bsName", src2 as "bsName")"},
	{ (REALproc) bsXOr1,  REALnoImplementation, "BitsXOr(source as "bsName")"},
	{ (REALproc) bsXOr2,  REALnoImplementation, "BitsXOr(src1 as "bsName", src2 as "bsName")"},
	{ (REALproc) bsDiff1, REALnoImplementation, "BitsDiff(source as "bsName")"},
	{ (REALproc) bsDiff2, REALnoImplementation, "BitsDiff(src1 as "bsName", src2 as "bsName")"},
	{ (REALproc) bsShift, REALnoImplementation, "BitsShift(shiftCount as Integer)"},
	{ (REALproc) bsBtClr, REALnoImplementation, "BitsClear(startBit as Integer, numBits as Integer)"},
	{ (REALproc) bsBtCp1, REALnoImplementation, "BitsCopy(srcBitNum as Integer, destBitNum as Integer, numBits as Integer)"},
	{ (REALproc) bsBtCp2, REALnoImplementation, "BitsCopy(source as "bsName", srcBitNum as Integer, destBitNum as Integer, numBits as Integer)"},
};

static REALclassDefinition bitsetClass = {
	kCurrentREALControlVersion,
	bsName,
	nil,
	sizeof(bitsetClassData),
	0,
	nil, // no default constructor here - we want to enforce using the one with the "bits" parameter
	(REALproc) bitsetClassDestructor,
	bitsetClassProperties,
	sizeof(bitsetClassProperties) / sizeof(REALproperty),
	bitsetClassMethods,
	sizeof(bitsetClassMethods) / sizeof(REALmethodDefinition),
	nil,
	0
};


//----------------
// Initialization
//----------------

void PluginEntry (void)
{
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
	
	REALRegisterClass (&bitsetClass);
}
