/*
 *  Copyright  1997-2002 Metrowerks Corporation.  All Rights Reserved.
 *
 *  Questions and comments to:
 *       <mailto:support@metrowerks.com>
 *       <http://www.metrowerks.com/>
 */

#include <iostream>

// this might be an implementation of CRC-8, using the polynomial x^8 + x^2 + x + 1.

static unsigned char CRC8(unsigned char inByte, unsigned char crc)
{
	for (int j=1; j <= 8; j++)
	{
		if ((crc ^ inByte) & 1)
			crc = ((crc ^ 0x18) >> 1) | 0x80;
		else
			crc = (crc >> 1) & 0x7f;
		inByte >>= 1;
	}
	return crc;
}

	static const unsigned char kElgato =   'E';	// which happens to be 0x45
	static const unsigned char kProProduct = 0x45;
	static const unsigned char kNMPProduct = 0x46;

class CodeGenerator
{
	public:
		enum { kCodeCount = 34 };
		
		CodeGenerator(unsigned char manufacturer,
					  unsigned char product);
		
		void DumpCodes();
		void TestOwnCorruptCodes();
		void TestOtherCodes(const CodeGenerator& otherGenerator);
		bool Acceptable(unsigned long code);
		
		unsigned long GetNthCode(int i) const;	// 0-based index
		int		CodeCount() const { return kCodeCount; }
		
		
	protected:
		void GenerateCodes();
		void TestOneCorruptCode(unsigned long testCode);
		void TestOneCode(unsigned long testCode);
		
	private:
		unsigned char mManufacturer;
		unsigned char mProduct;
		
		unsigned long mCodes[kCodeCount];
					 
};

int main()
{
	using namespace std;
	
	CodeGenerator codeGeneratorPro(kElgato, kProProduct);
	CodeGenerator codeGeneratorNMP(kElgato, kNMPProduct);
		
	codeGeneratorPro.DumpCodes();
	
	codeGeneratorPro.TestOwnCorruptCodes();
	
	codeGeneratorNMP.DumpCodes();
	codeGeneratorNMP.TestOwnCorruptCodes();
	
	cout << "testing Pro with NMP codes" << endl;
	codeGeneratorPro.TestOtherCodes(codeGeneratorNMP);
	
	cout << "testing NMP with Pro codes" << endl;
	codeGeneratorNMP.TestOtherCodes(codeGeneratorPro);
	
	cout << "all tests done" << endl;
	
	return 0;
}


CodeGenerator::CodeGenerator(unsigned char manufacturer,
					  unsigned char product)
					  : mManufacturer(manufacturer), mProduct(product)
{
	GenerateCodes();
}
					  
void CodeGenerator::GenerateCodes()
{
	for (unsigned char  key = 1; key <= kCodeCount; key++)
	{
		unsigned char crc = 0;
		crc = CRC8(mManufacturer, crc);
		crc = CRC8(mProduct, crc);
		crc = CRC8(key, crc);
		
		unsigned long keycode = (mManufacturer << 24) | (mProduct << 16) | (key << 8 | crc);
		
		mCodes[key-1] = keycode;
	}
}

void CodeGenerator::DumpCodes()
{
	using namespace std;
	
	cout << "codes for product " << hex << (int)mProduct << dec << endl;
	
	for (int key = 1; key <= kCodeCount; key++)
	{
		cout << "key ";
		cout.width(4);
		cout << key;
		cout.width(0);
		cout << "  " << hex << mCodes[key-1] << dec << endl;
	}
}

unsigned long CodeGenerator::GetNthCode(int i) const
{
	if (i < 0 or i >= kCodeCount)
		throw;
	return mCodes[i];
}

void CodeGenerator::TestOwnCorruptCodes()
{
	for (int i=0; i < kCodeCount; i++)
	{
		TestOneCorruptCode(mCodes[i]);	
	}
}

void CodeGenerator::TestOtherCodes(const CodeGenerator& otherGenerator)
{
	for (int i=0; i < otherGenerator.CodeCount(); i++)
	{
		unsigned long codeToTest = otherGenerator.GetNthCode(i);
		
		// neither of these two should produce any output
		TestOneCode(codeToTest);
		TestOneCorruptCode(codeToTest);
	}
}

void CodeGenerator::TestOneCode(unsigned long code)
{
	using namespace std;
	
	if (Acceptable(code))
		cout << "code " << hex << code << " is accepted for product " << (int)mProduct << endl;
}

void CodeGenerator::TestOneCorruptCode(unsigned long code)
{
	using namespace std;
	
	cout << "testing 1-bit errors for " << hex << code << dec << "...";
	
	for (int i=0; i < 32; i++)
	{
		unsigned long bitToFlip = 1UL << i;
		unsigned long falseCode = code ^ bitToFlip;
	
	
		if (Acceptable(falseCode))
			cout << "code " << hex << code << " is accepted when bit " << dec << i << "is flipped" << endl;
	}
	
	cout << "done, 2-bit...";
	
	
	for (int i = 0; i < 32; i++)
	{
		for (int j = 1; j < 32; j++)
		{
			unsigned long bitToFlip = (1UL << i) | (1UL << j);
			unsigned long falseCode = code ^ bitToFlip;
			
			if (Acceptable(falseCode))
				cout << "code " << hex << code << " is accepted when bits " << i << " and " << j << " are flipped" << endl;
		}
	}
	cout << "done, 3-bit...";
		
	for (int i = 0; i < 32; i++)
	{
		for (int j = 1; j < 32; j++)
		{
			for (int k = 2; k < 32; k++)
			{
				unsigned long bitToFlip = (1UL << i) | (1UL << j) | (1UL << k);
				unsigned long falseCode = code ^ bitToFlip;
			
				if (Acceptable(falseCode))
					cout << "code " << hex << code << " is accepted when bits " << i << ", " << j << " and " << k << " are flipped" << endl;
			}
		}
	}
	cout << "done" << endl;
}

bool CodeGenerator::Acceptable(unsigned long code)
{
	bool acceptable = false;
	if (((code >> 24) & 0xff) == mManufacturer)
	{
		if (((code >> 16) & 0xff) == mProduct)
		{
			unsigned int key = (code >> 8) & 0xff;
			
			if (key >= 1 and key <= kCodeCount)
			{
				unsigned int storedCRC = code & 0xff;
				unsigned int calcCRC = CRC8(mManufacturer, 0);
				calcCRC = CRC8(mProduct, calcCRC);
				calcCRC = CRC8(key, calcCRC);
				if (calcCRC == storedCRC)
					acceptable = true;
			}
		}
	}
	return acceptable;
}