/*
 * Plugin for REALbasic 4.5 and later to calculate CRCs
 *
 * This code is freeware, you may use and extend it as you like.
 *
 * The accompanying project file can be built with CodeWarrior Pro 6
 *
 * It was finished Jan 18, 2000 by Thomas Tempelmann.
 * Updated Apr 18, 2000: Added "...Cont" methods
 * Updated Mar 15, 2003: Update from CW Pro 4 to 6, added Carbon support.
 * Updated Apr 1, 2005: Added CRC-8 methods
 *
 * More RB-related stuff can be found here: <http://www.tempel.org/rb/>
 *
 * Enjoy and contribute!
 */

#include <string.h>
#include "rb_plugin.h"
#include "crc.h"



//-----------------------
// global plugin methods
//-----------------------

static long crc8CCITTPtr (Ptr p, long l)
{
	return crc8CCITT (p, l);
}

static long crc8CCITTcPtr (Ptr p, long l, long prev)
{
	return crc8CCITT (p, l, prev);
}

static long crc8CCITTStr (REALstring s)
{
	return crc8CCITT (s->CString(), s->Length());
}

static long crc8CCITTcStr (REALstring s, long prev)
{
	return crc8CCITT (s->CString(), s->Length(), prev);
}


static long crc16CCITTPtr (Ptr p, long l)
{
	return crc16CCITT (p, l);
}

static long crc16CCITTcPtr (Ptr p, long l, long prev)
{
	return crc16CCITT (p, l, prev);
}

static long crc16CCITTStr (REALstring s)
{
	return crc16CCITT (s->CString(), s->Length());
}

static long crc16CCITTcStr (REALstring s, long prev)
{
	return crc16CCITT (s->CString(), s->Length(), prev);
}


static long crc32Ptr (Ptr p, long l)
{
	return crc32 (p, l);
}

static long crc32cPtr (Ptr p, long l, long prev)
{
	return crc32 (p, l, prev);
}

static long crc32Str (REALstring s)
{
	return crc32 (s->CString(), s->Length());
}

static long crc32cStr (REALstring s, long prev)
{
	return crc32 (s->CString(), s->Length(), prev);
}


static REALstring crcDillonPtr (long bits, Ptr p, long l)
{
	unsigned long long crc = crcDillon (bits, p, l);
	return REALBuildString ((Ptr)&crc, sizeof(crc));
}

static REALstring crcDillonStr (long bits, REALstring s)
{
	unsigned long long crc = crcDillon (bits, s->CString(), s->Length());
	return REALBuildString ((Ptr)&crc, sizeof(crc));
}


static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) crc8CCITTPtr,	REALnoImplementation, "CRC_CCITT8InMem(address as Integer, length as Integer) as Integer" },
	{ (REALproc) crc8CCITTcPtr,	REALnoImplementation, "CRC_CCITT8InMemCont(address as Integer, length as Integer, prevCRC as Integer) as Integer" },
	{ (REALproc) crc8CCITTStr,	REALnoImplementation, "CRC_CCITT8OfStr(s as String) as Integer" },
	{ (REALproc) crc8CCITTcStr,	REALnoImplementation, "CRC_CCITT8OfStrCont(s as String, prevCRC as Integer) as Integer" },
	{ (REALproc) crc16CCITTPtr,	REALnoImplementation, "CRC_CCITTInMem(address as Integer, length as Integer) as Integer" },
	{ (REALproc) crc16CCITTcPtr,REALnoImplementation, "CRC_CCITTInMemCont(address as Integer, length as Integer, prevCRC as Integer) as Integer" },
	{ (REALproc) crc16CCITTStr,	REALnoImplementation, "CRC_CCITTOfStr(s as String) as Integer" },
	{ (REALproc) crc16CCITTcStr,REALnoImplementation, "CRC_CCITTOfStrCont(s as String, prevCRC as Integer) as Integer" },
	{ (REALproc) crc32Ptr,		REALnoImplementation, "CRC_32InMem(address as Integer, length as Integer) as Integer" },
	{ (REALproc) crc32cPtr,		REALnoImplementation, "CRC_32InMemCont(address as Integer, length as Integer, prevCRC as Integer) as Integer" },
	{ (REALproc) crc32Str,		REALnoImplementation, "CRC_32OfStr(s as String) as Integer" },
	{ (REALproc) crc32cStr,		REALnoImplementation, "CRC_32OfStrCont(s as String, prevCRC as Integer) as Integer" },
	{ (REALproc) crcDillonPtr,	REALnoImplementation, "CRC_DillonInMem(bitWidth as Integer, address as Integer, length as Integer) as String" },
	{ (REALproc) crcDillonStr,	REALnoImplementation, "CRC_DillonOfStr(bitWidth as Integer, s as String) as String" },
};


//------------------------------------------
// extension methods to "MemoryBlock" class
//------------------------------------------

static long crc8CCITTMemBlk (REALmemoryBlock mb, long ofs, long l)
{
	return crc8CCITT ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
}

static long crc8CCITTcMemBlk (REALmemoryBlock mb, long ofs, long l, long prev)
{
	return crc8CCITT ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l, prev);
}

static long crc16CCITTMemBlk (REALmemoryBlock mb, long ofs, long l)
{
	return crc16CCITT ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
}

static long crc16CCITTcMemBlk (REALmemoryBlock mb, long ofs, long l, long prev)
{
	return crc16CCITT ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l, prev);
}

static long crc32MemBlk (REALmemoryBlock mb, long ofs, long l)
{
	return crc32 ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
}

static long crc32cMemBlk (REALmemoryBlock mb, long ofs, long l, long prev)
{
	return crc32 ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l, prev);
}

static REALstring crcDillonMemBlk (REALmemoryBlock mb, long bits, long ofs, long l)
{
	unsigned long long crc = crcDillon (bits, (Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
	return REALBuildString ((Ptr)&crc, sizeof(crc));
}

static REALmethodDefinition memBlkClassMethods[] = {
	{ (REALproc) crc8CCITTMemBlk,	REALnoImplementation, "CRC_CCITT8(offset as Integer, numBytes as Integer) as Integer"},
	{ (REALproc) crc8CCITTcMemBlk,	REALnoImplementation, "CRC_CCITT8Cont(offset as Integer, numBytes as Integer, prevCRC as Integer) as Integer"},
	{ (REALproc) crc16CCITTMemBlk,	REALnoImplementation, "CRC_CCITT(offset as Integer, numBytes as Integer) as Integer"},
	{ (REALproc) crc16CCITTcMemBlk,	REALnoImplementation, "CRC_CCITTCont(offset as Integer, numBytes as Integer, prevCRC as Integer) as Integer"},
	{ (REALproc) crc32MemBlk,		REALnoImplementation, "CRC_32(offset as Integer, numBytes as Integer) as Integer"},
	{ (REALproc) crc32cMemBlk,		REALnoImplementation, "CRC_32Cont(offset as Integer, numBytes as Integer, prevCRC as Integer) as Integer"},
	{ (REALproc) crcDillonMemBlk,	REALnoImplementation, "CRC_Dillon(bitWidth as Integer, offset as Integer, numBytes as Integer) as String"},
};

static REALclassDefinition memBlkClass = {
	kCurrentREALControlVersion,
	"MemoryBlock", nil, 0, 0, (REALproc) nil, (REALproc) nil, nil, 0,
	memBlkClassMethods, sizeof(memBlkClassMethods) / sizeof(REALmethodDefinition),
};


//-----------------------
// plugin initialization
//-----------------------

void PluginEntry (void)
{
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
	
	REALRegisterClassExtension (&memBlkClass);
}
