/*
 * Sample-Plugin showing how to pass structs from and to REALbasic-Plugins
 *
 * This plugin actually is useful:
 *  It allows you to search for files with a certain file type and/or file creator.
 *  This search is very fast (uses the same low-level function as the Finder's
 *  Find File aka Sherlock uses). However, some File Systems do not support this
 *  fast search method (called CatSearch): On Floppy Disks and on CD-ROMs with
 *  a ISO 9660 format this function might not work (depending on the File System
 *  driver you've installed). On plain HFS/HFS+ volumes, it should always work.
 *
 * This plugin works only with MacOS, neither with Windows nor with Java.
 *
 * The enclosed project file requires CodeWarrior Pro 4 or later.
 * There's also a REALbasic Demo project that shows how to use this plugin.
 *
 * This code is freeware, you may use and extend it as you like.
 *
 * It was written on Feb 9, 99 by Thomas Tempelmann.
 * More RB-related stuff can be found here: <http://www.tempel.org/rb>
 *
 * Update by TT on 12 Feb 99:
 *   IsRemoteVolume() prevents crash by setting ioNamePtr to zero
 *
 * Update by TT on 8 Mar 99:
 *   Renamed the plugin from "CatalogSearch-Plugin" to "TT's CatSearch-Plugin".
 *   Added some more Catsearch-related functions.
 *   Moved Catsearch-unrelated functions into a new plugin (called "TT's FileMgr-Plugin").
 *
 * Update by TT on 22 Nov 99, v2.0
 *	 Added CatSearchSetName()
 *	 Improved VolSupportsCatSearch() to consider the gestaltFullExtFSDispatching flag
 *	 Uses now MoreFiles' PBCatSearchSyncCompat() to find files also on vols that
 *	  do not support CatSearch.
 *
 * Update by TT on 25 Nov 99, v2.0.1
 *	 Fixed the detection of whether a vol supports CatSearch. The wrong bit was tested.
 *
 * Update by TT on 11 Apr 00, v2.1
 *	 Updated for RB 2.1. Note that 68K code does not work with RB 2.0 any more.
 *
 * Update by TT on 24 Apr 01, v2.2
 *	 The project file now requires CW Pro 4 instead of Pro 3.
 *   CatSearchSetFileFlags() does not exclude Folders any more (SetFileType and SetCreator
 *    still do, however).
 *
 * Update by TT on 19 Jun 05, v2.2.1
 *	 Carbonized
 *   Tested on OS 10.4.1, found that VolSupportsCatSearch() does always return FALSE there;
 *    removed the 'hasFullExtFSDispatching' check to make it work on OS X. I must admit that
 *    I do not understand any more why this test was in there at all.
 *
 * Enjoy and contribute!
 */

#include <string.h>
#include "rb_plugin.h"

#include "Search.h"	// from MoreFiles (find it on ftp://ftp.apple.com/developer/Sample_Code/Files/)


// this is the struct where we store all our data in
typedef struct {
	long		check1;
	CSParam		pb;
	FSSpec		spec;
	CInfoPBRec	info1;
	CInfoPBRec	info2;
	Str27		volNameBuf;
	Str31		fileNameBuf;
	Handle		searchBufHdl;
	Boolean		volSupportsCS;
	long		check2;
} CatSearchRec;


// a prototype (forward declaration)
static Boolean volSupportsCS (short vRefNum);


static Handle CatSearchOpen (REALfolderItem volIn, Boolean allowRecursiveSearch)
// this creates the struct (allocates memory for it)
{
	FSSpec fsSpec;
	if (REALFSSpecFromFolderItem (&fsSpec, volIn)) {
		Boolean canUseCS = volSupportsCS(fsSpec.vRefNum);
		if (canUseCS || allowRecursiveSearch) {
			Handle h = NewHandle (sizeof(CatSearchRec));
			if (h && *h) {
				// init the CatSearch fields
				CatSearchRec *cs = (CatSearchRec*)*h;
				CSParam &pb = cs->pb;
				memset (cs, 0, sizeof (*cs));
				pb.ioVRefNum = fsSpec.vRefNum;
				pb.ioReqMatchCount = 1;
				pb.ioSearchTime = 15;	// return after about 1/60s of searching
				pb.ioOptBufSize = 0;
				cs->check1 = 'chk1';
				cs->check2 = 'chk2';
				cs->volSupportsCS = canUseCS;

				// invalidate the pointers in the pb so that noone tries to access them:
				pb.ioNamePtr = (StringPtr)0x41414141;
				pb.ioMatchPtr = (FSSpec*)0x41414141;
				pb.ioSearchInfo1 = (CInfoPBRec*)0x41414141;
				pb.ioSearchInfo2 = (CInfoPBRec*)0x41414141;
				// pb.ioOptBuffer = (Ptr)0x41414141;

				return h;
			}
		}
	}
	return nil;
}

static void CatSearchSetResponseTimeout (Handle h, long timeout)
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
			cs->pb.ioSearchTime = timeout;
		}
	}
}

static void CatSearchClose (Handle h)
// this releases the memory for the struct again
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
			DisposeHandle (h);
		}
	}
}

static long CatSearchSetFileType (Handle h, REALstring fileType)
// this sets some of the struct's values
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
			const unsigned char* ftStr = REALPString (fileType);
			OSType ftype = *(OSType*)&ftStr[1];
			cs->info1.hFileInfo.ioFlFndrInfo.fdType = ftype;
			cs->info2.hFileInfo.ioFlFndrInfo.fdType = (UInt32) -1;	// mask
			cs->info1.hFileInfo.ioFlAttrib &= ~ioDirMask;	// looking for files, not dirs
			cs->info2.hFileInfo.ioFlAttrib |= ioDirMask;	// set the mask for files/dirs
			cs->pb.ioSearchBits |= fsSBFlAttrib | fsSBFlFndrInfo;
			return 1;
		}
	}
	return 0;
}

static long CatSearchSetCreator (Handle h, REALstring creator)
// this sets some of the struct's values
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
			const unsigned char* cr = REALPString (creator);
			OSType fc = *(OSType*)&cr[1];
			cs->info1.hFileInfo.ioFlFndrInfo.fdCreator = fc;
			cs->info2.hFileInfo.ioFlFndrInfo.fdCreator = (UInt32) -1;	// mask
			cs->info1.hFileInfo.ioFlAttrib &= ~ioDirMask;	// looking for files, not dirs
			cs->info2.hFileInfo.ioFlAttrib |= ioDirMask;	// set the mask for files/dirs
			cs->pb.ioSearchBits |= fsSBFlAttrib | fsSBFlFndrInfo;
			return 1;
		}
	}
	return 0;
}

static long CatSearchSetName (Handle h, REALstring name, Boolean partial)
// This sets the file or dir name to look for.
// If partial is true, then all files that contain the name are found,
// otherwise only those that match the name exactly (case insensitively).
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
			const unsigned char* cr = REALPString (name);
			short n = cr[0];	// length of name
			if (n >= sizeof(cs->fileNameBuf)) {
				n = sizeof(cs->fileNameBuf) - 1;
			}
			BlockMoveData (cr, cs->fileNameBuf, n+1);
			cs->info1.hFileInfo.ioNamePtr = (StringPtr)0x41414141;	// just make it non-zero here - it'll be fixed later
			
			/* this would restrict the matching to folders only:
				cs->info1.hFileInfo.ioFlAttrib |= ioDirMask;	// looking for dirs, not files
				cs->info2.hFileInfo.ioFlAttrib |= ioDirMask;	// set the mask for files/dirs
				cs->pb.ioSearchBits |= fsSBFlAttrib;
			*/

			if (partial) {
				cs->pb.ioSearchBits |= fsSBPartialName;
			}
			return 1;
		}
	}
	return 0;
}

static long CatSearchSetFileFlags (Handle h, long mask, long selector)
// this adds specific File Flags to the Search Selection
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
			cs->info1.hFileInfo.ioFlFndrInfo.fdFlags = (cs->info1.hFileInfo.ioFlFndrInfo.fdFlags & ~mask) | (selector & mask);
			cs->info2.hFileInfo.ioFlFndrInfo.fdFlags |= mask;
/* TT 24 Apr 01: let's not limit it to files any more, as one may want to find invisible folders, too
			cs->info1.hFileInfo.ioFlAttrib &= ~ioDirMask;	// looking for files, not dirs
			cs->info2.hFileInfo.ioFlAttrib |= ioDirMask;	// set the mask for files/dirs
*/
			cs->pb.ioSearchBits |= fsSBFlAttrib | fsSBFlFndrInfo;
			return 1;
		}
	}
	return 0;
}

static long CatSearchNext (Handle h)
// this routine performs the real search action
// returns 1 if a match was found, 0 if none yet found (but timed out), negative if an err occured
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else {
		
			// since the CatSearchRec buffer might have been moved (it's not locked), we must
			// set the pointers right before they're accessed:
			CSParam &pb = cs->pb;
			pb.ioNamePtr = cs->volNameBuf;
			pb.ioMatchPtr = &cs->spec;
			pb.ioSearchInfo1 = &cs->info1;
			pb.ioSearchInfo2 = &cs->info2;
			if (pb.ioOptBuffer) {
				//pb.ioOptBuffer = cs->searchBuf;
			}
			if (cs->info1.hFileInfo.ioNamePtr) {
				cs->info1.hFileInfo.ioNamePtr = cs->fileNameBuf;
			}
			
			// now do the search
			OSErr err;
			if (cs->volSupportsCS) {
				err = PBCatSearchSync (&cs->pb);
			} else {
				err = IndexedSearch (&cs->pb, fsRtDirID);
			}

			// invalidate the pointers again so that we get an error right away when we access them accidentely:
			pb.ioNamePtr = (StringPtr)0x41414141;
			pb.ioMatchPtr = (FSSpec*)0x41414141;
			pb.ioSearchInfo1 = (CInfoPBRec*)0x41414141;
			pb.ioSearchInfo2 = (CInfoPBRec*)0x41414141;
			if (pb.ioOptBuffer) {
				pb.ioOptBuffer = (Ptr)0x41414141;
			}
			if (cs->info1.hFileInfo.ioNamePtr) {
				cs->info1.hFileInfo.ioNamePtr = (StringPtr)0x41414141;
			}

			if (err == catChangedErr) {
				return 2;
			} else if (!err) {
				return pb.ioActMatchCount == 0;	// returns 0 or 1
			}
			
			return err;
		}
	}
	return -1;
}

static REALfolderItem CatSearchGetItem (Handle h)
// this retrieves one of the struct's values
{
	if (h) {
		CatSearchRec *cs = (CatSearchRec*)*h;
		if (cs->check1 != 'chk1' || cs->check2 != 'chk2') {
			// DebugStr ("\pError in CatSearch-Plugin: bad handle passed");
		} else if (cs->pb.ioActMatchCount == 1) {
			return REALFolderItemFromFSSpec (&cs->spec);
		}
	}
	return NULL;
}


static Boolean volSupportsCS (short vRefNum)
{
	static Boolean fullExtFSDispatchingtested = false;
	static Boolean hasFullExtFSDispatching = false;

	// See if File Manager will pass CatSearch requests to external file systems
	if (!fullExtFSDispatchingtested) {
		long response;
		fullExtFSDispatchingtested = true;
		if (Gestalt(gestaltFSAttr, &response) == noErr) {
			hasFullExtFSDispatching = ((response & (1L << gestaltFullExtFSDispatching)) != 0);
		}
	}
	
//	if (hasFullExtFSDispatching) {	// !TT 19Jun05 removed this if() clause because otherwise we get always FALSE on OS X (Tiger)
		IOParam pb;
		GetVolParmsInfoBuffer buf;
		pb.ioVRefNum = vRefNum;
		pb.ioBuffer = (Ptr)&buf;
		pb.ioReqCount = 6;
		pb.ioNamePtr = nil;
		if (PBHGetVolParmsSync ((HParmBlkPtr)&pb) == noErr && pb.ioActCount >= 6) {
			if ((buf.vMAttrib & (1L<<bHasCatSearch)) != 0) return true;
		}
//	}

	return false;
}

static long VolSupportsCatSearch (REALfolderItem volIn)
{
	FSSpec fsSpec;
	if (REALFSSpecFromFolderItem (&fsSpec, volIn)) {
		return volSupportsCS (fsSpec.vRefNum);
	}
	return 0;
}

static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) CatSearchOpen, REALnoImplementation, "CatSearchOpen(volume as FolderItem, allowRecursiveSearch as Boolean) as Integer"},
	{ (REALproc) CatSearchSetResponseTimeout, REALnoImplementation, "CatSearchSetResponseTimeout(handle as Integer, milliSeconds as Integer)"},
	{ (REALproc) CatSearchClose, REALnoImplementation, "CatSearchClose(handle as Integer)"},
	{ (REALproc) CatSearchSetFileType, REALnoImplementation, "CatSearchSetFileType(handle as Integer, fileType as String) as Boolean"},
	{ (REALproc) CatSearchSetCreator, REALnoImplementation, "CatSearchSetCreator(handle as Integer, creator as String) as Boolean"},
	{ (REALproc) CatSearchSetName, REALnoImplementation, "CatSearchSetName(handle as Integer, name as String, partial as Boolean) as Boolean"},
	{ (REALproc) CatSearchNext, REALnoImplementation, "CatSearchNext(handle as Integer) as Integer"},
	{ (REALproc) CatSearchGetItem, REALnoImplementation, "CatSearchGetItem(handle as Integer) as FolderItem"},
	{ (REALproc) CatSearchSetFileFlags, REALnoImplementation, "CatSearchSetFileFlags(handle as Integer, mask as Integer, selector as Integer) as Boolean"},
	{ (REALproc) VolSupportsCatSearch, REALnoImplementation, "VolSupportsCatSearch(volume as FolderItem) as Boolean"},
};

void PluginEntry (void)
{
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
}
