/*
 * Plugin for REALbasic 2 that adds some new methods to RB's FolderItem class
 *
 * This code is freeware, you may use and extend it as you like.
 *
 * The accompanying project file was created using CodeWarrior Pro 5.
 *
 * It was written on July 31, 99 by Thomas Tempelmann.
 * Updated by TT on August 10, 99: Added two GetCreatorApp methods.
 * Updated by TT on March 22, 02: Added access to long names on HFS+ (not finished yet)
 * Updated by TT on April 8, 02: Launch renamed to TTsLaunch to avoid stupid compile conflict in RB 4.5a6
 *
 * More RB-related stuff can be found here: <http://www.tempel.org/rb/>
 *
 * Enjoy and contribute!
 */

#include <string.h>
#include <Memory.h>
#include "LaunchLib.h"		// this file is part of "TT-Libs", which can be downloaded from <http://www.tempel.org/macdev/#Libs>
#include "MoreDesktopMgr.h"	// this file is part of "MoreFiles", available somewhere on your CW CDs and at some places on the web.

#include "ResourceForkSanity.h"

#include "rb_plugin.h"

//----------------
// plugin classes
//----------------

static long openWith (REALfolderItem self, REALfolderItem app, long inFront)
{
	//DebugStr ("\pOpenWithApp");
	FSSpec selfSpec;
	OSErr err = -1;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		FSSpec appSpec;
		if (app == nil || REALFSSpecFromFolderItem (&appSpec, app)) {
			if (app != nil) {
				err = LaunchApplicationWithDocument (&appSpec, &selfSpec, !inFront);
			} else {
				err = OpenSpecifiedDocument (&selfSpec, !inFront);
			}
		}
	}
	return !err;
}

static long launch (REALfolderItem self, long inFront)
{
	return openWith (self, nil, inFront);
}

static REALfolderItem getCreatorSub (short vRefNum, OSType creator)
{
	FSSpec appSpec;
	if (FindApplicationByCreator (creator, vRefNum, &appSpec) == noErr) {
		return REALFolderItemFromFSSpec (&appSpec);
	}
	return nil;
}

static REALfolderItem getCreatorFromCode (REALfolderItem self, REALstring creator)
{
	FSSpec selfSpec;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		OSType fc = *(OSType*)creator->CString();
		return getCreatorSub (selfSpec.vRefNum, fc);
	}
	return nil;
}

static REALfolderItem getCreatorFromFile (REALfolderItem self)
{
	FSSpec selfSpec;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		FInfo fi;
		if (FSpGetFInfo (&selfSpec, &fi) == noErr) {
			return getCreatorSub (selfSpec.vRefNum, fi.fdCreator);
		}
	}
	return nil;
}


static REALpicture getIconFromTC (REALfolderItem self, long maxDepth, REALstring type, REALstring creator)
{
	FSSpec selfSpec;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		short iconType = kLarge8BitIcon;
		if (maxDepth < 8) {
			iconType = (maxDepth < 4) ? kLargeIcon : kLarge4BitIcon;
		}
		Boolean again;
		do {
			again = false;
			Handle h;
			if (noErr == DTGetIcon (nil, selfSpec.vRefNum, iconType, REALstringToOSType(creator), REALstringToOSType(type), &h)) {
				HLock (h);
				const char* src = *h;
				const short size = 32;
				REALpicture p;
				GWorldPtr world;
				Rect r;
				SetRect (&r, 0, 0, size, size);
				short depth = 8;
				if (iconType == kLarge4BitIcon) depth = 4;
				else if (iconType == kLargeIcon) depth = 1;
				NewGWorld (&world, depth, &r, nil, nil, 0);
				PixMapHandle pm = GetGWorldPixMap (world);
				Ptr dest = GetPixBaseAddr (pm);
				for (int i = 0; i < size; i++) {
					long n = (size/8)*depth;
					memcpy (dest, src, n);
					src += n;
					dest += (*pm)->rowBytes & 0x3FFF;
				}
				p = REALBuildPictureFromGWorld (world, true);
				HUnlock (h);
				DisposeHandle (h);
				return p;
			} else {
				if (iconType > kLargeIcon) {
					--iconType;
					again = true;
				}
			}
		} while (again);
	}
	return nil;
}

static long checkSanity (REALfolderItem self)
{
	Boolean sane = false;
	FSSpec selfSpec;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		CheckResourceForkSanity (&selfSpec, &sane);
	}
	return sane;
}

/*static REALstring getFullName (REALfolderItem self)
{
	FSSpec selfSpec, spec2;
	REALstring name = nil;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		...
		//OSErr err2 = FSpMakeFSRef (&selfSpec, &ref);
		//err2 = FSGetCatalogInfo (&ref, kFSCatInfoGettableInfo, &catInfo, &ucName, &spec2, &ref2);
	}
	return name;
}
*/

/*
static void setFullName (REALfolderItem self, REALstring name)
{
	FSSpec selfSpec;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		...
	}
}
*/

static REALmethodDefinition folderItemClassMethods[] = {
	{ (REALproc) openWith, REALnoImplementation, "OpenWithApp(app as FolderItem, inFront as Boolean) as Boolean"},
	{ (REALproc) launch, REALnoImplementation, "TTsLaunch(inFront as Boolean) as Boolean"},
	{ (REALproc) getCreatorFromFile, REALnoImplementation, "GetCreatorApp() as FolderItem"},
	{ (REALproc) getCreatorFromCode, REALnoImplementation, "GetCreatorApp(creatorCode as String) as FolderItem"},
	{ (REALproc) getIconFromTC, REALnoImplementation, "GetIconFromTC(maxDepth as Integer, typeCode as String, creatorCode as String) as Picture"},
	{ (REALproc) checkSanity, REALnoImplementation, "IsResourceSane() as Boolean"},
//	{ (REALproc) getFullName, (REALproc) setFullName, "FullName() as String"},
};

static REALclassDefinition folderItemClass = {
	kCurrentREALControlVersion,
	"FolderItem",
	nil,
	0,
	0,
	(REALproc) nil,
	(REALproc) nil,
	nil,
	0,
	folderItemClassMethods,
	sizeof(folderItemClassMethods) / sizeof(REALmethodDefinition),
	nil,
	0
};


void PluginEntry (void)
{
	REALRegisterClassExtension (&folderItemClass);
}
