/*
 * Plugin for REALbasic 2, providing the following classes:
 *
 *   FolderItem		(extends the existing class with new methods)
 *   TTsBinaryFork		(provides file access > 2GB)
 *
 * This code is freeware, you may use and extend it as you like.
 * However, I'd welcome it if you'd let me know what you did with it
 * and if you'd even grant me a copy of your work in case it's useful :-)
 *
 * The accompanying project file was created using CodeWarrior Pro 5.
 *
 * It was written on Oct 22, 99 by Thomas Tempelmann, rb@tempel.org
 * Updated May 26, 02 by TT: ptr compare against 0x2800 is now unsigned because OS X uses RAM at > 2GB
 *
 * More RB-related stuff can be found here: <http://www.tempel.org/rb/>
 *
 * Enjoy and contribute!
 */

#include <Files.h>
#include <Memory.h>
#include "rb_plugin.h"


#define	LargeFileClassName	"TTsBinaryFork"


//--------------------
//  Support routines
//--------------------

static Boolean	newAPIAvailable = false;

static Boolean SystemSupportsNewAPI ()
{
	long l;
	return (Gestalt (gestaltFSAttr, &l) == noErr) && (l & (1L<<gestaltHasHFSPlusAPIs)) != 0;
}

static Boolean VolSupportsNewAPINatively (short vRefNum)
{
	if (newAPIAvailable) {
		IOParam pb;
		GetVolParmsInfoBuffer buf;
		pb.ioVRefNum = vRefNum;
		pb.ioBuffer = (Ptr)&buf;
		pb.ioReqCount = sizeof(buf);
		pb.ioNamePtr = nil;
		if (PBHGetVolParmsSync ((HParmBlkPtr)&pb) == noErr && pb.ioActCount >= sizeof(buf)) {
			return (buf.vMExtendedAttributes & (1L<<bSupportsHFSPlusAPIs)) != 0;
		}
	}
	return false;
}

static Boolean VolSupports2TBFiles (short vRefNum)
{
	if (newAPIAvailable) {
		IOParam pb;
		GetVolParmsInfoBuffer buf;
		pb.ioVRefNum = vRefNum;
		pb.ioBuffer = (Ptr)&buf;
		pb.ioReqCount = sizeof(buf);
		pb.ioNamePtr = nil;
		if (PBHGetVolParmsSync ((HParmBlkPtr)&pb) == noErr && pb.ioActCount >= sizeof(buf)) {
			long m = (1L<<bSupports2TBFiles) | (1L<<bSupportsHFSPlusAPIs);
			if ((buf.vMExtendedAttributes & m) == m) {
				return true;
			}
		}
	}
	return false;
}

static void SimpleConvertUCToASCII (HFSUniStr255 src, Str255 dest)
// This is a quick-and-dirty Unicode-to-ASCII converter. It only works
// if the string only contains the basic ASCII char set (7 Bit).
{
	int n = src.length;
	StringPtr d = &dest[0];
	UniChar* s = &src.unicode[0];
	*d++ = n;
	while (n-- > 0) {
		*d++ = *s++;
	}
}

static void SimpleConvertASCIIToUC (const unsigned char* src, HFSUniStr255 &dest)
// This is a quick-and-dirty ASCII-to-Unicode converter. It only works
// correctly if the string only contains the basic ASCII char set (7 Bit).
{
	StringPtr s = (StringPtr) &src[0];
	int n = *s++;
	UniChar* d = &dest.unicode[0];
	dest.length = n;
	while (n-- > 0) {
		*d++ = *s++;
	}
}

// 
//------------------------
// LargeBinaryFile class
//------------------------


// --- the class data ---

extern REALclassDefinition LargeFileClass;	// forward declaration

struct LargeFileData
{
	// internal class properties
	Boolean ready;
	Boolean	wantsToWrite;
	Boolean	canWrite;
	Boolean	useNewAPI;
	SInt16	fRefNum;
};


// --- the class methods ---


static void LFClose (REALobject instance);	// forward declaration

static void LargeFileConstructor (REALobject instance, void *garbage)
{
	ClassData (LargeFileClass, instance, LargeFileData, me);

	me->ready = true;
}

static void LargeFileDestructor (REALobject instance, void *garbage)
{
	ClassData (LargeFileClass, instance, LargeFileData, me);
	
	LFClose (instance);
	
	me->ready = false;
}

static void LFSetup (REALobject instance, SInt16 theRefNum, Boolean useNewAPI, Boolean wantsToWrite)
{
	OSErr err;
	ClassData (LargeFileClass, instance, LargeFileData, me);
	
	me->fRefNum = theRefNum;
	me->useNewAPI = useNewAPI;
	me->wantsToWrite = wantsToWrite;
	
	// test if the file is actually writable
	if (useNewAPI) {
		FSForkInfo fi;
		err = FSGetForkCBInfo (theRefNum, 0, NULL, NULL, &fi, NULL, NULL);
		if (!err) {
			me->canWrite = (*(short*)&fi.flags & kioFCBWriteMask) != 0;
		}
	} else {
		FCBPBRec fcb;
		fcb.ioFCBIndx = 0;
		fcb.ioRefNum = theRefNum;
		fcb.ioNamePtr = NULL;
		err = PBGetFCBInfoSync (&fcb);
		if (!err) {
			me->canWrite = (fcb.ioFCBFlags & kioFCBWriteMask) != 0;
		}
	}
}

static void LFClose (REALobject instance)
{
	ClassData (LargeFileClass, instance, LargeFileData, me);

	if (me->fRefNum) {
		// close the file
		if (me->useNewAPI) {
			FSCloseFork (me->fRefNum);
		} else {
			FSClose (me->fRefNum);
		}
		me->fRefNum = 0;
	}
}

#if TARGET_CPU_68K
	typedef Float80 RBdouble;
#else
	typedef double RBdouble;
#endif

#if TARGET_CPU_68K
static void LFGetEOF (RBdouble* f, REALobject instance, long param)
#else
static RBdouble LFGetEOF (REALobject instance)
#endif
{
	OSErr err;
	ClassData (LargeFileClass, instance, LargeFileData, me);

	SInt64 ll;
	if (me->useNewAPI) {
		err = FSGetForkSize (me->fRefNum, &ll);
	} else {
		long l;
		err = GetEOF (me->fRefNum, &l);
		ll = l;
	}
	if (err) {
		ll = 0;
	}
	#if TARGET_CPU_68K
		*f = (double)ll;
	#else
		return (double)ll;
	#endif
}

static void LFSetEOF (REALobject instance, long param, RBdouble dbl)
{
	ClassData (LargeFileClass, instance, LargeFileData, me);

	if (me->useNewAPI) {
		FSSetForkSize (me->fRefNum, fsFromStart, dbl);
	} else {
		SInt64 ll = dbl;
		if (ll < 0x80000000) {
			SetEOF (me->fRefNum, (long)ll);
		}
	}
}

static long LFAtEOF (REALobject instance, long param)
{
	OSErr err;
	ClassData (LargeFileClass, instance, LargeFileData, me);

	if (me->useNewAPI) {
		FSForkInfo fi;
		err = FSGetForkCBInfo (me->fRefNum, 0, NULL, NULL, &fi, NULL, NULL);
		if (!err) {
			return fi.currentPosition >= fi.logicalEOF;
		}
	} else {
		FCBPBRec fcb;
		fcb.ioFCBIndx = 0;
		fcb.ioRefNum = me->fRefNum;
		fcb.ioNamePtr = NULL;
		err = PBGetFCBInfoSync (&fcb);
		if (!err) {
			return fcb.ioFCBCrPs >= fcb.ioFCBEOF;
		}
	}
	return true;
}

#if TARGET_CPU_68K
static void LFGetPos (RBdouble* f, REALobject instance, long param)
#else
static RBdouble LFGetPos (REALobject instance)
#endif
{
	OSErr err;
	ClassData (LargeFileClass, instance, LargeFileData, me);

	SInt64 ll;
	if (me->useNewAPI) {
		err = FSGetForkPosition (me->fRefNum, &ll);
	} else {
		long l;
		err = GetFPos (me->fRefNum, &l);
		ll = l;
	}
	if (err) {
		ll = 0;
	}
	#if TARGET_CPU_68K
		*f = (double)ll;
	#else
		return (double)ll;
	#endif
}

static void LFSetPos (REALobject instance, long param, RBdouble dbl)
{
	ClassData (LargeFileClass, instance, LargeFileData, me);

	if (me->useNewAPI) {
		FSSetForkPosition (me->fRefNum, fsFromStart, dbl);
	} else {
		SInt64 ll = dbl;
		if (ll < 0x80000000) {
			SetFPos (me->fRefNum, fsFromStart, (long)ll);
		}
	}
}

static Boolean LFReadRaw (REALobject instance, long amount, Ptr dataPtr, long noCache)
{
	OSErr err;
	ByteCount n;
	ClassData (LargeFileClass, instance, LargeFileData, me);
	
	if (amount <= 0 || (unsigned long)dataPtr <= 0x2800) {
		if (amount == 0) {
			return true;
		}
		// oops - illegal values!
		return false;
	}
	
	if (me->useNewAPI) {
		err = FSReadFork (me->fRefNum, fsAtMark, noCache?noCacheMask:0, amount, dataPtr, &n);
	} else {
		signed long sn = amount;
		err = FSRead (me->fRefNum, &sn, dataPtr);
		n = sn;
	}
	if (err) {
		n = 0;
	}
	return n == amount;
}

static Boolean LFWriteRaw (REALobject instance, long amount, Ptr dataPtr, long noCache)
{
	OSErr err;
	ByteCount n;
	ClassData (LargeFileClass, instance, LargeFileData, me);
	
	if (amount <= 0 || (unsigned long)dataPtr <= 0x2800) {
		if (amount == 0) {
			return true;
		}
		// oops - illegal values!
		return false;
	}
	
	if (me->useNewAPI) {
		err = FSWriteFork (me->fRefNum, fsAtMark, noCache?noCacheMask:0, amount, dataPtr, &n);
	} else {
		signed long sn = amount;
		err = FSWrite (me->fRefNum, &sn, dataPtr);
		n = sn;
	}
	if (err) {
		n = 0;
	}
	return n == amount;
}

static long LFCanWrite (REALobject instance, long param)
{
	ClassData (LargeFileClass, instance, LargeFileData, me);

	return (long) me->canWrite;
}


// --- the class description ---

static REALproperty LargeFileProperties[] = {
	{ nil, "CanWrite", "Boolean", 0, (REALproc) LFCanWrite },
	{ nil, "AtEOF", "Boolean", 0, (REALproc) LFAtEOF },
	{ nil, "Length", "Double", 0, (REALproc) LFGetEOF, (REALproc) LFSetEOF },
	{ nil, "Position", "Double", 0, (REALproc) LFGetPos, (REALproc) LFSetPos },
};

static REALmethodDefinition LargeFileMethods[] = {
	{ (REALproc) LFClose, REALnoImplementation, "Close()"},
	{ (REALproc) LFReadRaw, REALnoImplementation, "ReadBytes(amount as Integer, dataPtr as Integer, noNeedToCache as Boolean) as Boolean"},
	{ (REALproc) LFWriteRaw, REALnoImplementation, "WriteBytes(amount as Integer, dataPtr as Integer, noNeedToCache as Boolean) as Boolean"},
};

static REALclassDefinition LargeFileClass = {
	kCurrentREALControlVersion,
	LargeFileClassName,
	nil,
	sizeof(LargeFileData),
	0,
	(REALproc) LargeFileConstructor,
	(REALproc) LargeFileDestructor,
	LargeFileProperties,
	sizeof(LargeFileProperties) / sizeof(REALproperty),
	LargeFileMethods,
	sizeof(LargeFileMethods) / sizeof(REALmethodDefinition),
	nil,
	0
};



//--------------------------------
//  FolderItem (extension) class
//--------------------------------


// --- the class methods ---

static REALobject OpenLargeFile (REALfolderItem self, REALstring forkName, long write)
{
	OSErr err;
	FSSpec selfSpec;

	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		Boolean useNewAPI = VolSupportsNewAPINatively(selfSpec.vRefNum);
		SInt16 fRefNum;
		FSRef fRef;
		if (useNewAPI) {
			err = FSpMakeFSRef (&selfSpec, &fRef);
			if (!err) {
				// open the fork using the new HFS+ API which supports files > 2GB in size
				HFSUniStr255 fName;
				SimpleConvertASCIIToUC (forkName->PString(), fName);
				err = FSOpenFork (&fRef, fName.length, fName.unicode, write?fsCurPerm:fsRdPerm, &fRefNum);
			}
		} else {
			// open the fork using the classic API which does not support files > 2GB in size
			StringPtr fName = (StringPtr) forkName->PString();
			if (((short*)fName)[0] == 0x0144) {
				err = FSpOpenDF (&selfSpec, write?fsCurPerm:fsRdPerm, &fRefNum);
			} else if (((short*)fName)[0] == 0x0152) {
				err = FSpOpenRF (&selfSpec, write?fsCurPerm:fsRdPerm, &fRefNum);
			} else {
				err = errFSBadForkName;
			}
		}
		if (!err) {
			REALobject lf;
			lf = REALnewInstance (LargeFileClassName);
			if (lf) {
				// creating the instance succeeded - pass the file information to the class and then we're done here
				LFSetup (lf, fRefNum, useNewAPI, write != 0);
				return lf;
			}
			// creating the instance failed - close the file again
			if (useNewAPI) {
				FSCloseFork (fRefNum);
			} else {
				FSClose (fRefNum);
			}
		}
	}
	return nil;
}


static long CanHandleLarge (REALfolderItem self)
{
	FSSpec selfSpec;
	if (REALFSSpecFromFolderItem (&selfSpec, self)) {
		return VolSupports2TBFiles (selfSpec.vRefNum);
	}
	return false;
}



#if TARGET_CPU_68K
static void GetFreeSpace (RBdouble* f, REALfolderItem instance)
#else
static RBdouble GetFreeSpace (REALfolderItem instance)
#endif
{
	OSErr err;
	FSSpec spec;
	SInt64 ll = 0;
	if (REALFSSpecFromFolderItem (&spec, instance)) {
		long l = 0;
		if (Gestalt (gestaltFSAttr, &l) == noErr && (l & (1L << gestaltFSSupports2TBVols)) != 0) {
			XVolumeParam pb;
			pb.ioNamePtr = nil;
			pb.ioVolIndex = 0;
			pb.ioVRefNum = spec.vRefNum;
			err = PBXGetVolInfoSync (&pb);
			if (!err) {
				ll = pb.ioVFreeBytes;
			}
		} else {
			HVolumeParam pb;
			pb.ioNamePtr = nil;
			pb.ioVolIndex = 0;
			pb.ioVRefNum = spec.vRefNum;
			err = PBHGetVInfoSync ((HParmBlkPtr)&pb);
			if (!err) {
				ll = (unsigned long)pb.ioVFrBlk * (unsigned long)pb.ioVAlBlkSiz;
			}
		}
	}
	#if TARGET_CPU_68K
		*f = (double)ll;
	#else
		return (double)ll;
	#endif
}


// --- the class description ---

static REALmethodDefinition FolderItemClassMethods[] = {
	{ (REALproc) CanHandleLarge, REALnoImplementation, "VolumeSupportsLargeFiles() as Boolean"},
	{ (REALproc) GetFreeSpace, REALnoImplementation, "VolumeFreeSpace() as Double"},
	{ (REALproc) OpenLargeFile, REALnoImplementation, "OpenAsTTsBinaryFork(forkName as String, write as Boolean) as " LargeFileClassName},
};

static REALclassDefinition FolderItemClass = {
	kCurrentREALControlVersion,
	"FolderItem",
	nil,
	0,
	0,
	(REALproc) nil,
	(REALproc) nil,
	nil,
	0,
	FolderItemClassMethods,
	sizeof(FolderItemClassMethods) / sizeof(REALmethodDefinition),
	nil,
	0
};


//-----------------------
//  some global methods
//-----------------------


static Str255 gDFName, gRFName;

static REALstring GetDFName ()
{
	return REALBuildString ((Ptr)&gDFName[1], gDFName[0]);
}

static REALstring GetRFName ()
{
	return REALBuildString ((Ptr)&gRFName[1], gRFName[0]);
}

static long SysSupportsNewAPI ()
{
	return newAPIAvailable;
}

static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) GetRFName, REALnoImplementation, "SystemResourceForkName() as String"},
	{ (REALproc) GetDFName, REALnoImplementation, "SystemDataForkName() as String"},
	{ (REALproc) SysSupportsNewAPI, REALnoImplementation, "SystemSupportsLargeFiles() as Boolean"},
};


//--------------------------------------
//  main plugin entry (initialization)
//--------------------------------------

void PluginEntry (void)
{
	newAPIAvailable = SystemSupportsNewAPI ();
	if (newAPIAvailable) {
		HFSUniStr255 name;
		FSGetDataForkName (&name);
		SimpleConvertUCToASCII (name, gDFName);
		FSGetResourceForkName (&name);
		SimpleConvertUCToASCII (name, gRFName);
	} else {
		*(short*)&gDFName[0] = 0x0144;
		*(short*)&gRFName[0] = 0x0152;
	}
	
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}

	REALRegisterClass (&LargeFileClass);
	REALRegisterClassExtension (&FolderItemClass);
}
