/*
 * TT's MacFSSpec-Plugin
 *
 * Adds a "MacFSSpec" function (or read-only property) to the FolderItem class
 * Adds a "MacClassicName" function (or read-only property) to the FolderItem class
 *
 * Written by Thomas Tempelmann (rb@tempel.org), 21 Mar 2003
 */

#if TARGET_PPC || TARGET_CARBON
	#include <Script.h>
	#include <TextEncodingConverter.h>
#endif

#include "TT's Plugin Support.h"
#include "rb_plugin.h"


//----------------------------
// FolderItem class extension
//----------------------------

static REALmemoryBlock fsspec (REALfolderItem fi)
{
	#if TARGET_68K || TARGET_PPC || TARGET_CARBON
		REALmemoryBlock mb = REALNewMemoryBlock (sizeof(FSSpec));
		if (mb) {
			if (REALFSSpecFromFolderItem ((FSSpec*)REALMemoryBlockGetPtr(mb), fi)) {
				return mb;
			}
		}
	#endif
	return NULL;
}

/* Here's Apple's spec on the "fdScript" field in a catalog entry:
	fdScript 
		The script system for displaying the file's name. Ordinarily, the Finder (and
		the Standard File Package) displays the names of all desktop objects in the
		system script, which depends on the region-specific configuration of the system.
		The high bit of the byte in the fdScript field is set by default to 0, which
		causes the Finder to display the filename in the current system script. If the
		high bit is set to 1, the Finder (and the Standard File Package) displays the
		filename and directory name in the script whose code is recorded in the
		remaining 7 bits. 
*/

static REALstring fname (REALfolderItem fi)
{
	REALstring str = NULL;
	#if TARGET_68K || TARGET_PPC || TARGET_CARBON
		FSSpec spec;
		if (REALFSSpecFromFolderItem (&spec, fi)) {
			str = REALBuildString ((const char*)&spec.name[1], spec.name[0]);
			
			// we have the name, now we also need to get the script and
			// set the encoding accordingly.
			//
			// Note from TT: I am not sure if this works correctly. I have no
			//   ways to test this here right now. I've taken most of this code
			//   from MoreFilesX 1.0.1
			#if TARGET_PPC || TARGET_CARBON
			CInfoPBRec pb;
			pb.hFileInfo.ioVRefNum = spec.vRefNum;
			pb.hFileInfo.ioDirID = spec.parID;
			pb.hFileInfo.ioNamePtr = spec.name;
			pb.hFileInfo.ioFDirIndex = 0;
			OSErr err = PBGetCatInfoSync (&pb);
			if (!err) {
				TextEncoding textEncodingHint;
				ScriptCode	script;
				RegionCode	region;
				script = pb.hFileInfo.ioFlXFndrInfo.fdScript;
				if ((script & 0x80) == 0) {
					script = (ScriptCode)GetScriptManagerVariable (smSysScript);
				} else {
					script = script & 0x7F;
				}
				region = (RegionCode)GetScriptManagerVariable (smRegionCode);
				err = UpgradeScriptInfoToTextEncoding (script, kTextLanguageDontCare, region, NULL, &textEncodingHint);
				if (err == paramErr) {
					/* ok, ignore the region and try again */
					err = UpgradeScriptInfoToTextEncoding (script, kTextLanguageDontCare, kTextRegionDontCare, NULL, &textEncodingHint);
				}
				if (err) {
					/* ok... try something */
					textEncodingHint = kTextEncodingMacRoman;
				}
				TextEncodingBase enc = GetTextEncodingBase(textEncodingHint);
				REALSetStringEncoding (str, enc);
			}
			#endif
		}
	#endif
	return str;
}

static REALmethodDefinition folderClassMethods[] = {
	{ (REALproc) fsspec, REALnoImplementation, "MacFSSpec() as MemoryBlock"},
	{ (REALproc) fname, REALnoImplementation, "MacClassicName() as String"},
};

static REALclassDefinition folderExtClass = {
	kCurrentREALControlVersion,
	"FolderItem",
	nil,
	0, 0,
	nil, nil,
	nil, 0,
	folderClassMethods,
	sizeof(folderClassMethods) / sizeof(REALmethodDefinition),
	nil, 0
};


//----------------------------
//      Global methods
//----------------------------


static REALfolderItem fromSpec (REALmemoryBlock mb)
{
	#if TARGET_WIN32
		return nil;
	#else
		if (not mb) return nil;
		return REALFolderItemFromFSSpec((const FSSpec*)REALMemoryBlockGetPtr(mb));
	#endif
}

static long getVers ()
{
	return 0x010000;	// 1.0.0
}

static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) getVers, REALnoImplementation, "TTsMacFSSpecVersion() as Integer" },
	{ (REALproc) fromSpec, REALnoImplementation, "TTsFolderItemFromFSSpec(fsSpec as MemoryBlock) as FolderItem" },
};


// ---------------------------------
// ----- Plugin Initialization -----
// ---------------------------------

void PluginExit (void)
{
}

void PluginEntry (void)
{
	InitTTsPluginSupport ();
	
	REALRegisterClassExtension (&folderExtClass);

	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
}
