/*
 * Plugin for REALbasic 2.1.2 and later that adds some new methods to RB's MemoryBlock class
 *
 * This code is freeware, you may use and extend it as you like.
 *
 * The accompanying project file was created using CodeWarrior Pro 4, IDE 3.3, with
 * additional Carbon Headers from Apple, installed as a "Carbon Support" folder in
 * the same folder where the "MacOS Support" folder is located.
 *
 * More RB-related stuff can be found here: <http://www.tempel.org/rb/>
 *
 * Enjoy and contribute!
 *
 *
 * ---------
 *  History
 * ---------
 *
 *  June 7, 99
 *     Initially written by Thomas Tempelmann.
 *  July 31, 99
 *     Added "Address" method.
 *  Jan 18, 2000, v1.3
 *     Updated as a Windows (x86) plugin, and now using CW Pro 4 (formerly Pro 3).
 *  Jan 21, 2000, v1.3.1
 *     Bugfix: copyFewBytes now always takes a 32 bit instead of a 16 bit "len" parm,
 *     so that under Windows blocks > 32KB in size get copied properly.
 *  Mar 6, 2000, v1.4
 *     Added some code that checks the validity of the passed parameters on order
 *     to prevent bad effects, like crashes, when negative offsets and such are used.
 *     That means that some functions may now return different values:
 *       - Address() will return NIL if it is out of the MemBlk's range.
 *       - GetString will return NIL if the offset is out of range or 'length' is negative.
 *         It will return a sub-string if the offset is in range, but the length exceeds the
 *         MemBlk's size.
 *       - SetString will do nothing if the offset is out of range, 'length' is zero or
 *         negative.
 *         It will copy only a sub-string if the offset is in range, but the length exceeds
 *         the MemBlk's size, avoiding to modify data behind the MemBlk's range.
 *       - "Bool", "DoubleValue" and "OSType": If the value is not entirely inside the MemBlk:
 *         The getter functions return false or 0, and the setter functions do nothing.
 *       - "FSSpec": If the value is not entirely inside the MemBlk:
 *         The getter functions returns NIL, and the setter function does nothing.
 *  Jul 10, 2000, v1.5
 *     SwapBytes added (thanks to Franco Vaccari)
 *  Jul 14, 2000
 *     Int64Str added (thanks to Carsten Friehe for providing the code)
 *  Oct 21, 2000
 *     now uses memmove instead of my own copyBytes, which is apparently optimized
 *     for Mac and Windows
 *  Jan 25, 2001
 *     Added a global method "TTsMemoryBlockInit". Calling it from your BASIC code can prevent a
 *     bug in RB that would prevent this plugin not being included and causing crash in built
 *     apps. So, if you experience crashes using this plugin in your built apps, add the call to
 *     TTsMemoryBlockInit() somewhere in your app and see if this fixes the problem. Be aware that
 *     any other used plugins that provide only Classes may expose the same problem.
 *  Feb 16, 2001
 *     added TTsMemoryBlockPluginVersion(), MacPtrToMemoryBlock() and MacHandleToMemoryBlock()
 *  Feb 17, 2001, v1.6
 *     added Carbon support, released to public
 *  Mar 23, 2001, v1.7
 *     added BytesEqual()
 *  Sep 20, 2001, v1.8
 *     added ULong()
 *  Nov 11, 2001
 *     added FindBytes()
 *  Dec 16, 2002, v1.9
 *     renamed Size() into Length() to make RB 4's Size() method accessible (RB's can also set the size, while mine can't)
 *  Mar 23, 2003
 *     Removed several accessors entirely, because RB 3.5 added its own accessors.
 *     Here's the list of removed accessors and their RB 3.5 equivalents:
 *       Length: Size
 *       GetString, SetString: StringValue
 *       DoubleValue: DoubleValue
 *       Bool: BooleanValue
 */

#ifndef WIN32
	#include <MacMemory.h>
#else
	typedef char** Handle;
#endif

#include <stdio.h>
#include <string.h>

#include "rb_plugin.h"



//------------------
// global settings
//------------------

#define UseRB3_0orHigher 1	// set this to 0 to support RB 2.1.2


//------------------
// support routines
//------------------

static inline void copyBytes (char const * s0, Ptr d, unsigned long length)
{
	memmove ((void*)d, (void*)s0, length);
}

//----------------
// plugin classes
//----------------

extern REALclassDefinition memBlkClass;	// forward declaration

static inline Ptr getMemBlkPtr (REALobject instance)
// Returns the ptr to the actual data of a MemoryBlock
{
	#if UseRB3_0orHigher
		return (Ptr)REALMemoryBlockGetPtr((REALmemoryBlock)instance);
	#else
		return *(Ptr*)((Ptr)instance+0x1C);
	#endif
}

static inline long getSize (REALobject instance)
{
	// Attention: Memblk objects that were returned as a Ptr, either from a MemoryBlock,
	// or in parameters of "Declare" functions, always have a size of zero!
	#if UseRB3_0orHigher
		return REALMemoryBlockGetSize((REALmemoryBlock)instance);
	#else
		return *(long*)((Ptr)instance+0x18);
	#endif
}


static long address (REALobject instance, long offset)
{
	long size = getSize (instance);
	if (offset < 0 || (size && (offset > size))) return nil;
	Ptr memBlk = getMemBlkPtr (instance);
	return (long) (memBlk+offset);
}

static REALstring getString (REALobject instance, long offset, long length)
{
	long size = getSize (instance);
	if (length < 0 || offset < 0 || (size && (offset > size))) return nil;
	if (size && (offset + length > size)) length = size - offset;	// copies only the part that's inside the memblk
	Ptr memBlk = getMemBlkPtr (instance);
	return REALBuildString (memBlk+offset, length);
}

static void setString (REALobject instance, REALstring str, long offset)
{
	long size = getSize (instance);
	long length = str->Length();
	if (length <= 0 || offset < 0 || (size && (offset >= size))) return;
	if (size && (offset + length > size)) length = size - offset;	// copies only the part that's inside the memblk
	Ptr memBlk = getMemBlkPtr (instance);
	copyBytes (str->CString(), memBlk+offset, length);
}


static REALstring getOSType (REALobject instance, long offset)
{
	long size = getSize (instance);
	if (offset < 0 || (size && ((offset+4) > size))) return 0;
	Ptr memBlk = getMemBlkPtr (instance);
	return REALBuildString (memBlk+offset, 4);
}

static void setOSType (REALobject instance, long offset, REALstring v)
{
	long size = getSize (instance);
	if (offset < 0 || (size && ((offset+4) > size))) return;
	Ptr memBlk = getMemBlkPtr (instance);
	*(long*)(memBlk+offset) = *(long*)(v->CString());
}


static long getBool (REALobject instance, long offset)
{
	long size = getSize (instance);
	if (offset < 0 || (size && (offset >= size))) return false;
	Ptr memBlk = getMemBlkPtr (instance);
	return memBlk[offset] != 0;
}

static void setBool (REALobject instance, long offset, Boolean boolVal)
{
	long size = getSize (instance);
	if (offset < 0 || (size && (offset >= size))) return;
	Ptr memBlk = getMemBlkPtr (instance);
	memBlk[offset] = boolVal != 0;
}


#if TARGET_OS_MAC

	// FSSpec structs are not available under Windows
	
	static REALfolderItem getFSSpec (REALobject instance, long offset)
	{
		long size = getSize (instance);
		if (offset < 0 || (size && ((offset+sizeof(FSSpec)) > size))) return nil;
		Ptr memBlk = getMemBlkPtr (instance);
		FSSpec *spec = (FSSpec*)(memBlk+offset);
		return REALFolderItemFromFSSpec (spec);
	}

	static void setFSSpec (REALobject instance, long offset, REALfolderItem f)
	{
		long size = getSize (instance);
		if (offset < 0 || (size && ((offset+sizeof(FSSpec)) > size))) return;
		Ptr memBlk = getMemBlkPtr (instance);
		FSSpec *spec = (FSSpec*)(memBlk+offset);
		REALFSSpecFromFolderItem (spec, f);
	}

#else

	static REALfolderItem getFSSpec (REALobject instance, long offset)
	{
		return nil;
	}

	static void setFSSpec (REALobject instance, long offset, REALfolderItem f)
	{
	}

#endif


#if TARGET_CPU_68K
	typedef Float80 RBdouble;
#else
	typedef double RBdouble;
#endif

#if TARGET_CPU_68K
	static void getDouble (RBdouble* f, REALobject instance, long ofs)
	{
		long size = getSize (instance);
		if (ofs < 0 || (size && ((ofs+sizeof(RBdouble)) > size))) {
			*f = 0.0; // NAN would be even better here
			return;
		}
		Ptr memBlk = getMemBlkPtr (instance);
		*f = *(double*)(memBlk+ofs);
	}
#else
	static RBdouble getDouble (REALobject instance, long ofs)
	{
		long size = getSize (instance);
		if (ofs < 0 || (size && ((ofs+sizeof(RBdouble)) > size))) {
			return 0.0; // NAN would be even better here
		}
		Ptr memBlk = getMemBlkPtr (instance);
		return *(double*)(memBlk+ofs);
	}
#endif

static void setDouble (REALobject instance, long ofs, RBdouble f)
{
	long size = getSize (instance);
	if (ofs < 0 || (size && ((ofs+sizeof(RBdouble)) > size))) return;
	Ptr memBlk = getMemBlkPtr (instance);
	*(double*)(memBlk+ofs) = f;
}


#if TARGET_CPU_68K
	static void getUInt32 (RBdouble* f, REALobject instance, long ofs)
	{
		long size = getSize (instance);
		if (ofs < 0 || (size && ((ofs+sizeof(unsigned long)) > size))) {
			*f = 0.0; // NAN would be even better here
			return;
		}
		Ptr memBlk = getMemBlkPtr (instance);
		*f = (double)*(unsigned long*)(memBlk+ofs);
	}
#else
	static RBdouble getUInt32 (REALobject instance, long ofs)
	{
		long size = getSize (instance);
		if (ofs < 0 || (size && ((ofs+sizeof(unsigned long)) > size))) {
			return 0.0; // NAN would be even better here
		}
		Ptr memBlk = getMemBlkPtr (instance);
		unsigned long l = *(unsigned long*)(memBlk+ofs);
		return l;
	}
#endif

static void setUInt32 (REALobject instance, long ofs, RBdouble f)
{
	long size = getSize (instance);
	if (ofs < 0 || (size && ((ofs+sizeof(unsigned long)) > size))) return;
	Ptr memBlk = getMemBlkPtr (instance);
	*(unsigned long*)(memBlk+ofs) = (unsigned long)f;
}


static long findBytes0 (REALobject instance, long startOfs, long maxChunk, const char *targPtr, long targLen)
{
	const char *memBlk = getMemBlkPtr (instance);
	long memSize = getSize (instance);

	if (maxChunk == 0) maxChunk = memSize;
	if (memSize == 0) memSize = startOfs+maxChunk;
	
	if (targLen <= 0 or maxChunk <= 0 or startOfs < 0 or startOfs >= memSize) return -2; // stop this search entirely
	
	long endOfs = startOfs + maxChunk;
	if (endOfs > memSize) endOfs = memSize;
	const char *endPtr = memBlk + endOfs - targLen;
	
	const char *p = memBlk + startOfs;
	
	char firstByte = *targPtr++;
	
	do {
		if (*p == firstByte) {
			// check the rest
			const char *sp = p+1, *dp = targPtr;
			long len = targLen;
			do {
				if (--len == 0) return p-memBlk;	// found a match
			} while (*sp++ == *dp++);
		}
	} while (++p <= endPtr);

	return (endPtr < (memBlk+memSize)) ? -1 : -2; // not found, may continue (-1) or stop (-2)
}

static long findBytes2 (REALobject instance, long startOfs, long maxChunk, REALobject memBlk, long targOfs, long targLen)
{
	const char *targPtr = getMemBlkPtr (memBlk);
	long memSize = getSize (memBlk);
	if (targLen <= 0) targLen = memSize-targOfs;
	if (targLen <= 0 or targOfs < 0 or (memSize != 0 and targOfs >= memSize)) return -2;
	return findBytes0 (instance, startOfs, maxChunk, targPtr+targOfs, targLen);
}

static long findBytes1 (REALobject instance, long startOfs, long maxChunk, REALstring targ)
{
	long targLen = targ->Length();
	const char *targPtr = targ->CString();
	return findBytes0 (instance, startOfs, maxChunk, targPtr, targLen);
}

static void copyBytes1 (REALobject instance, long srcOfs, long length, long destOfs)
{
	Ptr memBlk = getMemBlkPtr (instance);
	copyBytes (memBlk+srcOfs, memBlk+destOfs, length);
}

static void copyBytes2 (REALobject instance, long srcOfs, long length, REALobject destMemBlk, long destOfs)
{
	Ptr memBlk = getMemBlkPtr (instance);
	Ptr memBlk2 = getMemBlkPtr (destMemBlk);
	copyBytes (memBlk+srcOfs, memBlk2+destOfs, length);
}

static long equBytes2 (REALobject instance, long srcOfs, long length, REALobject destMemBlk, long destOfs)
{
	Ptr memBlk = getMemBlkPtr (instance);
	Ptr memBlk2 = getMemBlkPtr (destMemBlk);
	return memcmp ((void*)(memBlk+srcOfs), (void*)(memBlk2+destOfs), length);
}

static void toMacPtr (REALobject instance, long offset, long length, Ptr dest)
{
	Ptr memBlk = getMemBlkPtr (instance);
	copyBytes (memBlk+offset, dest, length);
}

static void fromMacPtr (REALobject instance, Ptr src, long length, long destOfs)
{
	Ptr memBlk = getMemBlkPtr (instance);
	copyBytes (src, memBlk+destOfs, length);
}

static void toMacHdl (REALobject instance, long offset, long length, Handle dest)
{
	Ptr memBlk = getMemBlkPtr (instance);
	copyBytes (memBlk+offset, *dest, length);	// no need to Lock the Handle here, since this operation is atomic and does not move memory by itself
}

static void fromMacHdl (REALobject instance, Handle src, long length, long destOfs)
{
	Ptr memBlk = getMemBlkPtr (instance);
	copyBytes (*src, memBlk+destOfs, length);	// no need to Lock the Handle here, since this operation is atomic and does not move memory by itself
}

static void swapBytes (REALobject instance, long offset, long numBytes)
{
	Ptr memBlk = getMemBlkPtr (instance);
	Ptr p1, p2;
	long size = getSize (instance);
	if (numBytes <= 0 || offset < 0 || (size && (offset >= size))) return;
	if (size && (offset + numBytes > size)) numBytes = size - offset;
	p1 = memBlk + offset;
	p2 = p1 + numBytes;
	while (p1 < --p2) {
		char ch = *p1;
		*p1++ = *p2;
		*p2 = ch;
	}
}

static REALstring getSInt64 (REALobject instance, long offset)
{
	char output[32];
	long size = getSize (instance);
	if (offset < 0 || (size && ((offset+8) > size))) return nil;
	Ptr memBlk = getMemBlkPtr (instance);
	sprintf (output, "%lld", *(signed long long *)(memBlk+offset));
	return REALBuildString (output, strlen (output));
}

static void setSInt64 (REALobject instance, long offset, REALstring str)
{
	long size = getSize (instance);
	if (offset < 0 || (size && ((offset+8) > size))) {
		// Error: offset out of range!
		return;
	}
	Ptr memBlk = getMemBlkPtr (instance);
	sscanf (str->CString(), "%lld", memBlk+offset);
}


#if not UseRB3_0orHigher
static REALproperty memBlkClassProperties[] =
{
	{ nil, "Length", "Integer", 0, (REALproc) getSize },
};
#endif

static REALmethodDefinition memBlkClassMethods[] =
{
	#if not UseRB3_0orHigher
	{ (REALproc) getDouble,  (REALproc) setDouble, "DoubleValue(offset as Integer) as Double"},
	{ (REALproc) getBool,    (REALproc) setBool,   "Bool(offset as Integer) as Boolean"},
	#endif

	{ (REALproc) getString,  REALnoImplementation, "GetString(offset as Integer, numBytes as Integer) as String"},
	{ (REALproc) setString,  REALnoImplementation, "SetString(str as String, offset as Integer)"},

	{ (REALproc) address,    REALnoImplementation, "Address(offset as Integer) as Integer"},
	{ (REALproc) getOSType,  (REALproc) setOSType, "OSType(offset as Integer) as String"},
	{ (REALproc) getUInt32,  (REALproc) setUInt32, "ULong(offset as Integer) as Double"},
	{ (REALproc) getSInt64,  (REALproc) setSInt64, "Int64Str(offset as Integer) as String"},

	{ (REALproc) swapBytes,  REALnoImplementation, "SwapBytes(offset as Integer, numBytes as Integer)"},

	{ (REALproc) equBytes2,  REALnoImplementation, "BytesEqual(srcOfs as Integer, numBytes as Integer, destBlk as MemoryBlock, destOfs as Integer) as Boolean"},

	{ (REALproc) copyBytes1, REALnoImplementation, "CopyBytes(srcOfs as Integer, numBytes as Integer, destOfs as Integer)"},
	{ (REALproc) copyBytes2, REALnoImplementation, "CopyBytes(srcOfs as Integer, numBytes as Integer, destBlk as MemoryBlock, destOfs as Integer)"},

	{ (REALproc) toMacPtr,   REALnoImplementation, "CopyBytesToMacPtr(srcOfs as Integer, numBytes as Integer, destPtr as Integer)"},
	{ (REALproc) fromMacPtr, REALnoImplementation, "CopyBytesFromMacPtr(srcPtr as Integer, numBytes as Integer, destOfs as Integer)"},
	{ (REALproc) toMacHdl,   REALnoImplementation, "CopyBytesToMacHandle(srcOfs as Integer, numBytes as Integer, destHandle as Integer)"},
	{ (REALproc) fromMacHdl, REALnoImplementation, "CopyBytesFromMacHandle(srcHandle as Integer, numBytes as Integer, destOfs as Integer)"},

	{ (REALproc) findBytes1, REALnoImplementation, "FindString(srcOfs as Integer, maxBytes as Integer, target as String) as Integer"},
	{ (REALproc) findBytes2, REALnoImplementation, "FindBytes(srcOfs as Integer, maxBytes as Integer, target as MemoryBlock, targOfs as Integer, targLen as Integer) as Integer"},

	{ (REALproc) getFSSpec, (REALproc) setFSSpec, "FSSpec(offset as Integer) as FolderItem"},
};

static REALclassDefinition memBlkClass = {
	kCurrentREALControlVersion,
	"MemoryBlock",
	nil,
	0,
	0,
	(REALproc) nil,
	(REALproc) nil,
	#if UseRB3_0orHigher
	nil, 0,
	#else
	memBlkClassProperties, sizeof(memBlkClassProperties) / sizeof(REALproperty),
	#endif
	memBlkClassMethods, sizeof(memBlkClassMethods) / sizeof(REALmethodDefinition),
	nil, 0
};


#if UseRB3_0orHigher

	// REALPtrToMemoryBlock() is not available in RB 2.1.2 and before:

	static REALmemoryBlock ptrToMB (Ptr p)
	{
		return REALPtrToMemoryBlock (p);
	};

	static REALmemoryBlock hdlToMB (Handle h)
	{
		if (!h or !*h) return nil;
		#if not WIN32
			HLock (h);
		#endif
		return REALPtrToMemoryBlock (*h);
	};

#endif

static long getVers ()
{
	return 0x010900;
}

static void init ()
{
	// nothing to do here
}

static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) getVers, REALnoImplementation, "TTsMemoryBlockVersion() as Integer" },
	{ (REALproc) init, REALnoImplementation, "TTsMemoryBlockInit()"},
	#if UseRB3_0orHigher
		{ (REALproc) ptrToMB, REALnoImplementation, "MacPtrToMemoryBlock(thePtr as Integer) as MemoryBlock"},
		{ (REALproc) hdlToMB, REALnoImplementation, "MacHandleToMemoryBlock(theHdl as Integer) as MemoryBlock"},
	#endif
};

void PluginEntry (void)
{
	REALRegisterClassExtension (&memBlkClass);

	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
}
