/*
 * Serial I/O plugin for RB 3.0 and later
 *
 * By Thomas Tempelmann, http://www.tempel.org/rb/
 */

#if TARGET_OS_WIN32
	#include <MacTypes.h>
	static void inline DebugStr(unsigned char* p) {}
	#define log(x) OutputDebugString(x)
#else
	#include <MacMemory.h>
	#include "dcon.h"
	#define log(x) dprintf(x)
#endif

#if TARGET_CPU_68K
	#include <A4Stuff.h>
#endif

// Std C
#include <string.h>

// Mac OS
#if TARGET_OS_MAC
	#include <Devices.h>
	#include <Serial.h>
	#include "SerialConfig.h"
#endif

// RB
#include "TT's Plugin Support.h"
#include "rb_plugin.h"


typedef long	RBinteger;
typedef Boolean	RBboolVar;	// use this for Boolean variables stored in this plugin
typedef Boolean	RBboolPar;	// use this to receive Boolean parameters (both by value and by reference)
typedef long	RBboolRet;	// use this when returning a Boolean as a function result



// -- RB plugin specific code --


typedef struct ListCarrier {
	ListCarrier	*prev, *next;
	REALobject	obj;
} ListCarrier;

ListCarrier	gRoot;

static void listAttach (ListCarrier &l, REALobject instance)
{
	ListCarrier &r = gRoot;
	l.next = r.prev->next;
	l.prev = r.prev;
	r.prev->next = &l;
	r.prev = &l;
	l.obj = instance;
}

static void listDetach (ListCarrier &l)
{
	l.prev->next = l.next;
	l.next->prev = l.prev;
}


extern REALclassDefinition serClass;	// forward declaration

struct serClassData
{
	ListCarrier		link;
	RBinteger		lastError;
	#if TARGET_OS_MAC
	IOParam			in, out;
	SerialSettings	conf;
	REALstring		inBufStr;
//	RBinteger		bytesAvail;
	#endif
};

static void serClassConstructor (REALobject instance)
{
	ClassData (serClass, instance, serClassData, me);
	listAttach (me->link, instance);
}

static void serClassDestructor (REALobject instance)
{
	ClassData (serClass, instance, serClassData, me);
	listDetach (me->link);
	#if TARGET_OS_MAC
	SerialClose (&me->in, &me->out, false);
	REALUnlockString (me->inBufStr);
	me->inBufStr = nil;
	#endif
}

static RBboolRet serOpen (REALobject instance, RBinteger portIdx, RBinteger baud, RBinteger bufSize)
{
	ClassData (serClass, instance, serClassData, me);
	#if TARGET_OS_MAC
	me->conf.baudrate = baud;
	me->conf.stopbits = 0;
	me->conf.parity = 0;
	me->conf.databits = 8;
	me->conf.handshake = 0;
	me->lastError = SerialOpen (&me->in, &me->out, portIdx, me->conf);
	if (not me->lastError) {
		REALUnlockString (me->inBufStr);
		me->inBufStr = REALBuildString (nil, bufSize);
		SerSetBuf (me->in.ioRefNum, (Ptr)me->inBufStr->CString(), bufSize);
	}
	#endif
	return not me->lastError;
}

static void close (REALobject instance)
{
	ClassData (serClass, instance, serClassData, me);
	#if TARGET_OS_MAC
	SerialClose (&me->in, &me->out, false);
	REALUnlockString (me->inBufStr);
	me->inBufStr = nil;
	#endif
	me->lastError = 0;
}

static REALstring readStr (REALobject instance, RBinteger amount)
{
	ClassData (serClass, instance, serClassData, me);
	REALstring s = nil;
	#if TARGET_OS_MAC
	if (amount > 0) {
		s = REALBuildString (nil, amount);
		me->in.ioBuffer = (Ptr)s->CString();
		me->in.ioReqCount = amount;
		me->lastError = PBReadSync ((ParmBlkPtr)&me->in);
		long actCnt = me->in.ioActCount;
		if (amount != actCnt) {
			// make a new string with the correct size
			REALstring s2 = REALBuildString (nil, actCnt);
			memcpy ((Ptr)s2->CString(), s->CString(), actCnt);
			REALUnlockString (s);
			s = s2;
		}
	} else {
		me->lastError = 0;
	}
	#endif
	return s;
}

static RBinteger readMem (REALobject instance, REALmemoryBlock mb, RBinteger amount)
{
	ClassData (serClass, instance, serClassData, me);
	long actCnt = 0;
	#if TARGET_OS_MAC
	long size = REALMemoryBlockGetSize (mb);
	if (size != 0 and (amount < 0 or amount > size)) amount = size;
	if (amount != 0) {
		me->in.ioBuffer = (Ptr)REALMemoryBlockGetPtr(mb);
		me->in.ioReqCount = amount;
		me->lastError = PBReadSync ((ParmBlkPtr)&me->in);
		actCnt = me->in.ioActCount;
	} else {
		me->lastError = 0;
	}
	#endif
	return actCnt;
}

static RBinteger writeStr (REALobject instance, REALstring str)
{
	ClassData (serClass, instance, serClassData, me);
	long amount = str->Length();
	long actCnt = 0;
	#if TARGET_OS_MAC
	if (amount > 0) {
		me->out.ioBuffer = (Ptr)str->CString();
		me->out.ioReqCount = amount;
		me->lastError = PBWriteSync ((ParmBlkPtr)&me->out);
		actCnt = me->out.ioActCount;
	} else {
		me->lastError = 0;
	}
	#endif
	return actCnt;
}

static RBinteger writeMem (REALobject instance, REALmemoryBlock mb, RBinteger amount)
{
	ClassData (serClass, instance, serClassData, me);
	long size = REALMemoryBlockGetSize (mb);
	if (size != 0 and (amount < 0 or amount > size)) amount = size;
	long actCnt = 0;
	#if TARGET_OS_MAC
	if (amount != 0) {
		me->out.ioBuffer = (Ptr)REALMemoryBlockGetPtr(mb);
		me->out.ioReqCount = amount;
		me->lastError = PBWriteSync ((ParmBlkPtr)&me->out);
		actCnt = me->out.ioActCount;
	} else {
		me->lastError = 0;
	}
	#endif
	return actCnt;
}

static RBinteger avail (REALobject instance)
{
	ClassData (serClass, instance, serClassData, me);
	long l = 0;
	#if TARGET_OS_MAC
		me->lastError = SerGetBuf (me->in.ioRefNum, &l);
	#endif
	return l;
}

static REALmethodDefinition serClassMethods[] = {
	{ (REALproc) serOpen, REALnoImplementation, "Open(portIdx as Integer, baud as Integer, bufSize as Integer) as Boolean"},
	{ (REALproc) close, REALnoImplementation, "Close()"},
	{ (REALproc) avail, REALnoImplementation, "Available() as Integer"},
	{ (REALproc) readStr, REALnoImplementation, "Read(byteCnt as Integer) as String"},
	{ (REALproc) readMem, REALnoImplementation, "Read(buffer as MemoryBlock, byteCnt as Integer) as Integer"},
	{ (REALproc) writeStr, REALnoImplementation, "Write(data as String) as Integer"},
	{ (REALproc) writeMem, REALnoImplementation, "Write(data as MemoryBlock, byteCnt as Integer) as Integer"},
};

static REALproperty serClassProperties[] = {
	{ nil, "Error", "Integer", 0, REALstandardGetter, REALnoImplementation, FieldOffset(serClassData, lastError) },
//	{ nil, "BytesAvailable", "Integer", 0, REALstandardGetter, REALnoImplementation, FieldOffset(serClassData, bytesAvail) },
};

static REALclassDefinition serClass = {
	kCurrentREALControlVersion,
	"TTsSerial", nil,
	sizeof(serClassData), 0,
	(REALproc) serClassConstructor, (REALproc) serClassDestructor,
	serClassProperties,
	sizeof(serClassProperties) / sizeof(REALproperty),
	serClassMethods,
	sizeof(serClassMethods) / sizeof(REALmethodDefinition),
	nil, 0
};


// -- plugin init/exit --

void PluginExit (void)
{
	// call left over constructors (if RB's IDE kills a running app, it may leave objects behind)
	ListCarrier *r = &gRoot;
	ListCarrier *l = r->next;
	while (l != r) {
		serClassDestructor (l->obj);
		l = l->next;
	}
}

void PluginEntry (void)
{
	InitTTsPluginSupport ();

	gRoot.prev = &gRoot;
	gRoot.next = &gRoot;

	REALRegisterClass (&serClass);

	#if TARGET_API_MAC_OS8
		#if TARGET_CPU_68K
			SetCurrentA4();
		#endif
	#endif
}
