/*
 * Plugin for REALbasic 4.5 and later (no 68K support)
 *
 * By Thomas Tempelmann, <http://www.tempel.org/rb/>
 *
 *  07 Jul 02	(TT)	First version
 *  22 Mar 03	(TT)	RemoveEncoding sets enc to 0xffff instead of 0x0600 now;
 *           	    	TTsUTF8toISO_8859_1 sets string encoding now to ISO Latin-1
 *           	    	and made length values signed so that they'd not get into
 *           	    	infinite loops if the source string is incomplete.
 *   1 Apr 03	(TT)	TTsUTF8toISO_8859_1 does not set encoding any more because
 *           	    	it causes some performance problems which I like to avoid.
 */

#ifndef WIN32
	#include <MacMemory.h>
#else
	typedef char** Handle;
	typedef unsigned char UInt8;
	typedef signed char SInt8;
	typedef unsigned long UInt32;
	typedef long SInt32;
#endif

#include <stdio.h>
#include <string.h>

#include "rb_plugin.h"


//-----------------------------------------------------------------------
// begin of some UTF-8 functions
//

/*
 * Part of this code was take from Theo's String Stuff Plugin (as of Aug 8, 2002)
 */

/*
 * Index into the table below with the first byte of a UTF-8 sequence to
 * get the number of trailing bytes that are supposed to follow it.
 */
static const char trailingBytesForUTF8[256] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

/*
 * Magic values subtracted from a buffer value during UTF-8 conversion.
 * This table contains as many values as there might be trailing bytes
 * in a UTF-8 sequence.
 */
static const UInt32 offsetsFromUTF8[] = { 0x00000000UL, 0x00003080UL, 0x000E2080UL, 0x03C82080UL };

#define UNI_REPLACEMENT_CHAR (UInt32)0x0000FFFD
#define UNI_MAX_BMP (UInt32)0x0000FFFF
#define UNI_MAX_UTF16 (UInt32)0x0010FFFF
#define UNI_MAX_UTF32 (UInt32)0x7FFFFFFF

static void convertUTF8toISO_8859_1 (UInt8* srcIn, SInt32 lenIn, UInt8* dstIn, SInt32 dstLen)
// assuming that the UTF-8 sequence was built solely from an ISO-8859-1 encoded string,
// we will now convert it back, by moving the chars in place, returning the same or
// a shorter string.
{
	UInt8* source = srcIn;
	UInt8* srcEnd = srcIn + lenIn;
	UInt8* target = dstIn;
	UInt8* dstEnd = dstIn + dstLen;
	
	while (source < srcEnd and target < dstEnd) {
		UInt8 ch2 = *source;
		if ((SInt8)ch2 < 0) {
			int extraBytesToRead = trailingBytesForUTF8[ch2];
			UInt32 ch = 0;
			switch (extraBytesToRead) {
				default: extraBytesToRead = 3;
				case 3:	ch += *source++; ch <<= 6;
				case 2:	ch += *source++; ch <<= 6;
				case 1:	ch += *source++; ch <<= 6;
				case 0:	ch += *source++;
			}
			ch -= offsetsFromUTF8[extraBytesToRead];
			*target++ = ch;
		} else {
			*target++ = ch2;
			++source;
		}
	}
}

static Boolean needsConvertUTF8toISO_8859_1 (UInt8* srcIn, SInt32 lenIn, SInt32 &lenOut)
// just like ConvertUTF8toISO_8859_1, but will only prescan and return the
// length the dest string would require, and false if no conversion is necessary
{
	SInt32 len = lenIn, len2 = 0;
	Boolean needsConv = false;
	UInt8* source = srcIn;
	
	while (len-- > 0) {
		UInt8 ch2 = *source++;
		if ((SInt8)ch2 < 0) {
			needsConv = true;
			int extraBytesToRead = trailingBytesForUTF8[ch2];
			if (extraBytesToRead > 3) extraBytesToRead = 3;
			source += extraBytesToRead;
			len -= extraBytesToRead;
		}
		++len2;
	}
	lenOut = len2;
	return needsConv;
}

//
// end of some UTF-8 functions
//-----------------------------------------------------------------------


static REALstring utf8toISO (REALstring str)
{
	SInt32 len;
	REALstring destStr = nil;
	if (str) {
		if (not needsConvertUTF8toISO_8859_1 ((UInt8*)str->CString(), str->Length(), len)) {
			destStr = str;
			REALLockString (str);
		} else {
			destStr = REALBuildString (nil, len);
			convertUTF8toISO_8859_1 ((UInt8*)str->CString(), str->Length(), (UInt8*)destStr->CString(), len);
		}
	}
	// no, I must not do this, or performance in MyCom goes way down: if (destStr) REALSetStringEncoding (destStr, 0x0201);
	return destStr;
}

static void remEnc (REALstring str)
{
	REALSetStringEncoding (str, 0xFFFF);
}

static void setEnc (REALstring str, long enc)
{
	REALSetStringEncoding (str, enc);
}

static long getEnc (REALstring str)
{
	return REALGetStringEncoding (str);
}

static long getVers ()
{
	return 0x010100;	// v1.1.0
}

static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) getVers, REALnoImplementation, "TTsStringVersion() as Integer" },
	{ (REALproc) setEnc, REALnoImplementation, "TTsStringSetEncoding(s as String, encoding as Integer)"},
	{ (REALproc) getEnc, REALnoImplementation, "TTsStringGetEncoding(s as String) as Integer"},
	{ (REALproc) remEnc, REALnoImplementation, "TTsStringRemoveEncoding(s as String)"},
	{ (REALproc) utf8toISO, REALnoImplementation, "TTsUTF8toISO_8859_1(utf8string as String) as String"},
};

//------------------

void PluginEntry (void)
{
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
}

// EOF
