/*
 * UDF Socket Class for Windows
 *
 * Written and Copyright by Thomas Tempelmann, tt@tempel.org
 *
 * History:
 *	July 25, 01: Initial version.
 */


#include <string.h>
#include <stdlib.h>
#if TARGET_OS_WIN32
	#define WIN32_LEAN_AND_MEAN
//	#include <winsock2.h>
	static void inline DebugStr(unsigned char* p) {}
	#define log(x) OutputDebugString(x)
	#define PrepareCallback() 
	#define EnterCallback()
	#define ExitCallback()
#else
	#if TARGET_CPU_68K
		#include <SetupA4.h>
	#endif
	#include <MacMemory.h>
	#include "dcon.h"
	#define log(x) dprintf(x)
#endif

// Note: the following includes should come last, after all other includes:

#include "TT's Plugin Support.h"
#include "rb_plugin.h"


//------------------------
// Basic type definitions
//------------------------

typedef long RBinteger;
typedef Boolean	RBboolVar;	// use this for Boolean variables stored in this plugin
typedef Boolean	RBboolPar;	// use this to receive Boolean parameters (both by value and by reference)
typedef long	RBboolRet;	// use this when returning a Boolean as a function result




//---------------------
// Socket support code
//---------------------

#define DEFAULT_PORT 5001
//#define DEFAULT_PROTO SOCK_STREAM // TCP
#define DEFAULT_PROTO SOCK_DGRAM // UDP

#if TARGET_OS_WIN32

static SOCKET msgsock = 0;

static long sock_open ()
{
	unsigned short port = DEFAULT_PORT;
	int socket_type = DEFAULT_PROTO;
	struct sockaddr_in local;
	WSADATA wsaData;
	SOCKET listen_socket;
	int retval;

	if ((retval = WSAStartup(0x202,&wsaData)) != 0) {
		//fprintf(stderr,"WSAStartup failed with error %d\n",retval);
		WSACleanup();
		return -1;
	}

	local.sin_family = AF_INET;
	local.sin_addr.s_addr = INADDR_ANY;
	local.sin_port = htons(port);
	listen_socket = socket(AF_INET, socket_type,0); // TCP socket

	if (listen_socket == INVALID_SOCKET){
		//fprintf(stderr,"socket() failed with error %d\n",WSAGetLastError());
		WSACleanup();
		return -2;
	}

	if (bind(listen_socket,(struct sockaddr*)&local,sizeof(local) ) == SOCKET_ERROR) {
		//fprintf(stderr,"bind() failed with error %d\n",WSAGetLastError());
		closesocket (listen_socket);
		WSACleanup();
		return -3;
	}
	
	msgsock = listen_socket;
	
	return 0;
}

static long sock_read ()
{
	char Buffer[1600];
	int retval;
	int fromlen;
	struct sockaddr_in from;
	Boolean dataAvail = false;

	// check if data is available
	{
		struct timeval timeout = {0,0};
		fd_set set;
		FD_ZERO (&set);
        FD_SET (msgsock, &set);
		if (select (msgsock+1, &set, 0, &set, &timeout) > 0) {
			dataAvail = true;
		}
	}

	if (dataAvail) {
		fromlen = sizeof (from);
		retval = recvfrom (msgsock, Buffer, sizeof (Buffer), 0, (struct sockaddr*)&from, &fromlen);
		//printf("Received %d bytes, data [%s] from client\n",retval,Buffer);
	} else {
		retval = 0;
	}
	
	return retval;
}

static long sock_close ()
{
	if (msgsock) closesocket (msgsock);
	msgsock = 0;
	WSACleanup();
	return 0;
}

#else

	// not Windows:

	static long sock_open () { return 0; }
	static long sock_read () { return 0; }
	static long sock_send () { return 0; }
	static long sock_close () { return 0; }

#endif

//-----------------------------
// Socket Class implementation
//-----------------------------


extern REALclassDefinition socketClass;	// forward declaration


struct socketClassData
{
	long dummy;
};


static void socketClassConstructor (REALobject instance)
{
	ClassData (socketClass, instance, socketClassData, me);
}

static void socketClassDestructor (REALobject instance)
{
	ClassData (socketClass, instance, socketClassData, me);
	sock_close ();
}

static RBinteger socketIntGetter (REALobject instance, long param)
{
	ClassData (socketClass, instance, socketClassData, me);
	return 0; //me->aInteger;
}

static void socketIntSetter (REALobject instance, long param, RBinteger value)
{
	ClassData (socketClass, instance, socketClassData, me);
//	me->aInteger = value;
}


typedef RBinteger (*SocketEventProc) (REALobject theObjInst, RBinteger someValue);	// this is the C version of the "SocketEvent" declaration below, but with an added obj reference parm

static REALevent socketClassEvents[] = {
	{ "SocketEvent(someValue as Integer) as Integer" },
};

static RBinteger callSocketEvent (REALobject instance, RBinteger firstArgument)
{
	RBinteger result;
	
	int idx = 0;			// this index selects from the socketClassEvents array. 0 is the first entry.
	SocketEventProc fp = (SocketEventProc) REALGetEventInstance ((REALcontrolInstance)instance, &socketClassEvents[idx]);

	if (fp) {
		result = fp (instance, firstArgument);
	} else {
		result = 0;
	}
	
	return result;
}

static long callEvent (REALobject instance, RBinteger aValue)
{
	ClassData (socketClass, instance, socketClassData, me);
	return callSocketEvent (instance, aValue);
}


static long doOpen (REALobject instance)
{
	ClassData (socketClass, instance, socketClassData, me);
	return sock_open ();
}

static long doRead (REALobject instance)
{
	ClassData (socketClass, instance, socketClassData, me);
	return sock_read ();
}

static long doSend (REALobject instance)
{
	ClassData (socketClass, instance, socketClassData, me);
	return 0; // sock_send ();
}

static void doClose (REALobject instance)
{
	ClassData (socketClass, instance, socketClassData, me);
	sock_close ();
}


static REALmethodDefinition socketClassMethods[] = {
	{ (REALproc) callEvent,	REALnoImplementation, "CallEvent(aValue as Integer) as Integer"},
	{ (REALproc) doOpen,	REALnoImplementation, "Open() as Integer"},
	{ (REALproc) doRead,	REALnoImplementation, "Read() as Integer"},
	{ (REALproc) doSend,	REALnoImplementation, "Send() as Integer"},
	{ (REALproc) doClose,	REALnoImplementation, "Close()"},
};

static REALproperty socketClassProperties[] = {
	{ nil, "SocketInt", "Integer",         0, (REALproc) socketIntGetter,    (REALproc) socketIntSetter,    0 /* param */ },
};


static REALclassDefinition socketClass = {
	kCurrentREALControlVersion,
	"TTsUDFSocket",
	nil,
	sizeof(socketClassData),
	0,
	(REALproc) socketClassConstructor, (REALproc) socketClassDestructor,
	socketClassProperties,
	sizeof(socketClassProperties) / sizeof(REALproperty),
	socketClassMethods,
	sizeof(socketClassMethods) / sizeof(REALmethodDefinition),
	socketClassEvents,
	sizeof(socketClassEvents) / sizeof(REALevent),
	nil, 0
};


// ---------------------------------
// ----- Plugin Initialization -----
// ---------------------------------

	
void PluginInit (void)
{
}

void PluginExit (void)
{
}

void PluginEntry (void)
{
	InitTTsPluginSupport ();

	#if TARGET_CPU_68K
		PrepareCallback ();
	#endif
	
	REALRegisterClass (&socketClass);
}
