/*
 * plugin to monitor window events under Windows
 *
 * Made 01Nov01 by Thomas Tempelmann with major help from James Milne
 *
 * the purpose of this plugin was to catch the "quit" event of an MDI app
 * in order to make the test app "CloseHideTest2.rbp" work when you close
 * the app via either of these methods under Windows OS:
 * - top right close box of MDI frame
 * - Alt-F4 (or click in top left corner of MDI frame, then select "close")
 * - "End Task" from the Task Mgr (ctrl-alt-delete)
 *
 * This, however, can not be achieved with this plugin. Testing shows that
 * we get some of these messages before the RB app processes them, but
 * not all. Plus, we never get a clear "quit" msg, but rather msgs such as
 * a click on a control, and we would still have to analyse these msgs to
 * tell whether they'd lead to a quit of the app.
 *
 * I make this code available to other who want to monitor window events
 * in Windows, however.
 */

-> see Mail from Alfred v.H. on Nov 9, 01: he explained that I get
the WM_CLOSED only if I use a globally installed handler, not only
from a RB class. Plus, I must then NOT modify the GWL_WNDPROC before
calling the old handler.

Also, James Milne says that I can indeed call Events from here.

//--------------------
// Your includes here
//--------------------

#include <string.h>
#if TARGET_OS_WIN32
	static void inline DebugStr(unsigned char* p) {}
	#define log(x) OutputDebugString(x)
	#define PrepareCallback() 
	#define EnterCallback()
	#define ExitCallback()
#else
	#include <MacMemory.h>
	#include "dcon.h"
	#define log(x) dprintf(x)
#endif

//-------------------------
// RB header includes here
//-------------------------

#include "TT's Plugin Support.h"
#include "rb_plugin.h"


//-----------------------
// Some type definitions
//-----------------------

typedef long RBinteger;
typedef Boolean	RBboolVar;	// use this for Boolean variables stored in this plugin
typedef Boolean	RBboolPar;	// use this to receive Boolean parameters (both by value and by reference)
typedef long	RBboolRet;	// use this when returning a Boolean as a function result


//-----------------------
// Add global RB methods
//-----------------------

//-----------------------
// the following code comes from James Milne, 31Oct01, slightly modified by TT:

#if not TARGET_OS_WIN32
	typedef long HWND;
	typedef long WNDPROC;
	typedef long LONG;
#endif

struct EventProcessorData {
    REALobject instance;
    HWND hWnd;
    LONG    oldWindowProc;
    LONG    oldUserData;
    char	quitting;
};

extern REALclassDefinition eventProcessorObj;

static RBboolRet quitState (REALobject instance)
{
    ClassData(eventProcessorObj, instance, EventProcessorData, data);
	return data->quitting;
}

static void
EventProcessor_Constructor(REALobject instance)
{
    ClassData(eventProcessorObj, instance, EventProcessorData, data);
    data->instance = instance;
}

static void
EventProcessor_Destructor(REALobject instance)
{
#if TARGET_OS_WIN32
    ClassData(eventProcessorObj, instance, EventProcessorData, data);
    if(data->oldWindowProc) {
        SetWindowLong(data->hWnd, GWL_WNDPROC, data->oldWindowProc);
    }
    if(data->oldUserData) {
        SetWindowLong(data->hWnd, GWL_USERDATA, data->oldUserData);
    }
#endif
}

#if TARGET_OS_WIN32

static LRESULT CALLBACK
ProcessorWndProc(HWND hWnd, UINT message, UINT wParam, LONG lParam)
{
    EventProcessorData *data = (EventProcessorData*) GetWindowLong(hWnd, GWL_USERDATA);

    // Check the message
    switch(message) {
    case WM_DESTROY:

		data->quitting = true;
		log("WM_DESTROY\n");
		
		// TT believes that calling an Event from here is not a good idea and will probably crash the RB runtime:
        //void (*pQuitEvent)(REALobject instance) = nil;
        //pQuitEvent = REALGetEventInstance(data->instance, &EventProcessorEvents[0]);
        //if(pQuitEvent) pQuitEvent(data->instance);

        break;
        
    case 15:
    case 32:
    case 132:
    case 512:
    	// some mouse movement and other uninteresting stuff
    	break;
    	
    default:
    	
		char s[80];
		sprintf (s, "msg: %ld, wPar: %ld, lPar: %ld\n", message, wParam, lParam);
		log (s);
    }

    // Pass the message on to the old window proc, reinstating
    // the previous user data before you call it.
    LRESULT result;
    LONG curUserData, curWindowProc;

    curUserData = GetWindowLong(hWnd, GWL_USERDATA);
    curWindowProc = GetWindowLong(hWnd, GWL_WNDPROC);

    SetWindowLong(hWnd, GWL_USERDATA, data->oldUserData);
    SetWindowLong(hWnd, GWL_WNDPROC, data->oldWindowProc);

    result = CallWindowProc ((WNDPROC)data->oldWindowProc, hWnd, message, wParam, lParam);

    SetWindowLong(hWnd, GWL_USERDATA, curUserData);
    SetWindowLong(hWnd, GWL_WNDPROC, curWindowProc);

    return result;
}

#endif

static void
EventProcessor_SetWindow(REALobject instance, HWND hWnd, RBboolPar forParent)
{
#if TARGET_OS_WIN32
    ClassData(eventProcessorObj, instance, EventProcessorData, data);

	// James Milne says:
    // This method doesn't clean up right if you already have a window
    // set for the object. However, I couldn't be arsed finishing it
    // correctly.

	if (forParent) {
		hWnd = GetParent (hWnd);
	}
	
    data->hWnd = hWnd;
    if(hWnd) {
        data->oldWindowProc = GetWindowLong(data->hWnd, GWL_WNDPROC);
        data->oldUserData = GetWindowLong(data->hWnd, GWL_USERDATA);
        SetWindowLong(data->hWnd, GWL_WNDPROC, (LONG)ProcessorWndProc);
        SetWindowLong(data->hWnd, GWL_USERDATA, (LONG)data);
    }
#endif
}


REALmethodDefinition EventProcessorMethods[] = {
    { (REALproc) EventProcessor_SetWindow, nil, "SetWindow(hWnd as Integer, forParent as Boolean)" },
	{ (REALproc) quitState, REALnoImplementation, "Quitting() as Boolean" },
};

REALclassDefinition eventProcessorObj = {
    kCurrentREALControlVersion,
    "Win32EventMonitor", nil,
    sizeof(EventProcessorData), 0,
    (REALproc)EventProcessor_Constructor,
    (REALproc)EventProcessor_Destructor,
    nil, 0,
    EventProcessorMethods,
    sizeof(EventProcessorMethods) / sizeof(REALmethodDefinition),
};



// ---------------------------------
// ----- Plugin Initialization -----
// ---------------------------------
	
void PluginInit (void)
{
}

void PluginExit (void)
{
}

void PluginEntry (void)
{
	InitTTsPluginSupport ();

	REALRegisterClass (&eventProcessorObj);
}
