/*

	Expat Realbasic Plugin
	
	Original author: Doug Holton, July 10, 2000, doug.holton@vanderbilt.edu
	Improved by: Thomas Tempelmann, rb@tempel.org

	v1.0.6, updated Jan 6, 04 by Thomas Tempelmann (rb@tempel.org):
		prepared for use by RS in RB 5.5:
		Using SDK 5.5a11 now;
		Removed use of "QuitHandler.cp" and "TT's Plugin Support.cp";
		Using unmodified version of PluginMain.cpp, removed __initialize and __terminate calls
		in Linker settings for PPC and Carbon targets along with it;
		Renamed "SetEncoding" to "SetInputEncoding" to make it more clear that this does not
		change the encoding of the strings the parser passes to the Events but only specifies
		how the input is to be interpreted;
		Performed a memory leak test and fixed leaks in UnknownEntity and UnknownEncoding
		Potential problems and To Do items:
			- a better sample prj should be made, perhaps one that re-creates a valid XML file,
			  and an example should include the use of DTDs.
			- while the Attributes could be stored in a Dictionary, this would slow down runtime
			  performance compared to the current technique using a separate class to access them.
			- the parser output is always UTF-8. This might not be wanted in Linux.
			- the parser output converts line separators into CRs. This may not be wanted in Linux and Windows
			- the encoding property of input strings is ignored - instead the encoding arg in the XML file
			  is inspected or "SetInputEncoding" must be set accordingly. But the parser only supports the
			  following input encodings: ISO-8859-1, US-ASCII, UTF-8, UTF-16.

	updated Apr 29, 02 by Thomas Tempelmann (rb@tempel.org): added x86 output, used latest expat sources,
		exchanged the plugin support with mine, fixed crash caused by call of XML_ParserFree() with invalid parser reference,
		renamed property accessor from "AmParsing" to "Parsing", changed attr parm so that it uses a class to access the items,
		disabled the "Alphabetize multiple attribute tags", renamed Class to "TTsXMLParser", renamed a few methods and
		parms to better match the Java SAX Parser names.
		Named v1.0.2

	updated Apr 30, 02 by Thomas Tempelmann:
		Comments and other text get their LFs converted to CRs now, too.
		Fixed a leak (the Attributes were left allocated)
		Named v1.0.3

	updated Apr 30, 03 by Thomas Tempelmann:
		Made "Encoding" a function to avoid a crash in RB 5.1 when mistakenly calling the prop's getter,
		  and renamed it to "SetEncoding"
		Named v1.0.5

	Please see the readme for expat (expat.html) for more info.
	Below I copy the excerpt about the license:
	
		expat - XML Parser Toolkit

		Version 19990307

		Copyright (c) 1998, 1999 James Clark.  Expat is subject to the Mozilla Public License Version 1.0.  You may
		as a special exception elect to use the GNU
		General Public License.  Please contact me if you wish to
		negotiate an alternative license.

		Expat is an XML 1.0 parser
		written in C. It aims to be fully conforming.  It is currently not a
		validating XML processor.  The current production version of expat can
		be downloaded from ftp://ftp.jclark.com/pub/xml/expat.zip.

 */

//Includes:
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "xmlparse.h"
#include "codepage.h"
#include "xmlfile.h"
#include "xmltchar.h"

#include "rb_plugin.h"

//#include "dcon.h"

#ifdef XML_UNICODE_WCHAR_T
#define XML_T(x) L ## x
#else
#define XML_T(x) x
#endif


#define ParserPluginName "TTsXMLParser"
#define AttrPluginName "TTsXMLParserAttributes"


extern REALclassDefinition xmlattrclass;

struct xmlattrclassData {
	const char	**atts;
	long		count;
};

extern REALclassDefinition xmlparserclass;

struct xmlparserclassData {
	XML_Parser  parser;
	long		depth;
	Boolean		amparsing;
};

REALevent xmlparserEvents[] = {
	{ "CharacterData(data as String)"}, //0
	{ "StartTag(name as String, atts as "AttrPluginName")"}, //1
	{ "EndTag(name as String)"}, //2
	{ "Error(errCode as Integer, errText as String)"}, //3
	{ "Comment(data as String)" }, //4
	{ "UnknownEncoding(name as String) as Integer"}, //5
	{ "DefaultHandler(data as String)"}, //6
	{ "ProcessingInstruction(target as String, data as String)"}, //7
	{ "StartDocument()"}, //8
	{ "EndDocument()"}, //9
	{ "StartCData()"}, //10
	{ "EndCData()"}, //11
	{ "UnknownEntity(name as String) as Boolean"}, //12
};



///////CLASS METHODS//////////////

static const long encUTF8 = 0x08000100;

static void startElement(void *userData, const char *name, const char **atts)
{
	int		nAtts=0;
	int		offset=0;
	int		l=0;
	XML_Char **p;

	REALcontrolInstance instance = (REALcontrolInstance)userData;
	void (*fp)(REALcontrolInstance, REALstring, REALobject);
	fp = (void (*)(REALcontrolInstance, REALstring, REALobject)) REALGetEventInstance(instance, &xmlparserEvents[1]);
	if (fp) {

		//First get the number of key-val pairs and the length of all the strings together.
		p = (char **)atts;
		while (*p) ++p;
		nAtts = (p - atts) >> 1;

		// create the Attributes object
		REALobject attrObj = REALnewInstance (AttrPluginName);
		ClassData (xmlattrclass, attrObj, xmlattrclassData, attrData);
		attrData->count = nAtts;
		attrData->atts = atts;

		//Fire the start element event:
		REALstring str = REALBuildString (name,strlen(name));
		REALSetStringEncoding (str, encUTF8);
		fp (instance, str, attrObj);
		REALUnlockString(str);
	
		REALUnlockObject (attrObj);
	}
}

static void endElement(void *userData, const char *name)
{

	REALcontrolInstance instance = (REALcontrolInstance)userData;	
	void (*fp)(REALcontrolInstance, REALstring);
	fp = (void (*)(REALcontrolInstance, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[2]);
	if (fp) {
		REALstring str = REALBuildString(name,strlen(name));
		REALSetStringEncoding (str, encUTF8);
		fp (instance, str);
		REALUnlockString(str);
	}
}

static void convertLFtoCR (REALstring str)
{
	// convert all LFs into CRs (the parser has already converted any CRs into LFs before passing the text to us):
	char* s2 = (char*) str->CString();
	int len = str->Length();
	for (int i = 0; i < len; ++i) {
		if (*s2 == 10) *s2 = 13;
		++s2;
	}
}

static void characterData(void *userData, const XML_Char *s, int len)
{
	REALcontrolInstance instance = (REALcontrolInstance)userData;	
	void (*fp)(REALcontrolInstance, REALstring);
	fp = (void (*)(REALcontrolInstance, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[0]);
	if (fp) {
		REALstring str = REALBuildString(s,len);
		REALSetStringEncoding (str, encUTF8);
		convertLFtoCR (str);
		fp (instance, str);
		REALUnlockString(str);
	}
}

static void metaComment(XML_Parser parser, const XML_Char *data)
{
	REALcontrolInstance instance = (REALcontrolInstance) ((void *)(parser));
	void (*fp)(REALcontrolInstance, REALstring);
	fp = (void (*)(REALcontrolInstance, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[4]);
	if (fp) {
		REALstring str = REALBuildString(data,strlen(data));
		REALSetStringEncoding (str, encUTF8);
		convertLFtoCR (str);
		fp (instance, str);
		REALUnlockString(str);
	}
}

static int unknownEntity(void *userData, const XML_Char *name)
{
	REALcontrolInstance instance = (REALcontrolInstance) userData;
	int (*fp)(REALcontrolInstance, REALstring);
	fp = (int (*)(REALcontrolInstance, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[12]);
	if (fp) {
		REALstring str = REALBuildString(name,strlen(name));
		REALSetStringEncoding (str, encUTF8);
		int res = fp (instance, str);
		REALUnlockString(str);
		return res;
	}
	return 0;
}


/*
typedef struct {
  int map[256];
  void *data;
  int (*convert)(void *data, const char *s);
  void (*release)(void *data);
} XML_Encoding;
*/
static int unknownEncoding(void *userData,
		    const XML_Char *name,
		    XML_Encoding *info)
{
	REALcontrolInstance instance = (REALcontrolInstance) userData;
	int (*fp)(REALcontrolInstance, REALstring);
	fp = (int (*)(REALcontrolInstance, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[5]);
	if (fp) {
		REALstring str = REALBuildString(name,strlen(name));
		REALSetStringEncoding (str, encUTF8);
		int res = fp (instance, str);
		REALUnlockString(str);
		return res;
	}
	return 0;
}

static void markup(XML_Parser parser, const XML_Char *s, int len)
{
	//REALcontrolInstance instance = (REALcontrolInstance) XML_GetUserData(parser);
	REALcontrolInstance instance = (REALcontrolInstance) ((void *)(parser));
	void (*fp)(REALcontrolInstance, REALstring);
	fp = (void (*)(REALcontrolInstance, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[6]);
	if (fp) {
		REALstring str = REALBuildString(s,len);
		REALSetStringEncoding (str, encUTF8);
		convertLFtoCR (str);
		fp (instance, str);
		REALUnlockString(str);
	}
}

static void processingInstruction(void *userData, const XML_Char *target, const XML_Char *data)
{
	REALcontrolInstance instance = (REALcontrolInstance) userData;
	void (*fp)(REALcontrolInstance, REALstring, REALstring);
	fp = (void (*)(REALcontrolInstance, REALstring, REALstring)) REALGetEventInstance(instance, &xmlparserEvents[7]);
	if (fp) {
		REALstring str1 = REALBuildString(target,strlen(target));
		REALstring str2 = REALBuildString(data,strlen(data));
		REALSetStringEncoding (str1, encUTF8);
		REALSetStringEncoding (str2, encUTF8);
		convertLFtoCR (str2);
		fp (instance, str1, str2);
		REALUnlockString(str1);
		REALUnlockString(str2);
	}
}

static void metaStartCdataSection(XML_Parser parser)
{
	//REALcontrolInstance instance = (REALcontrolInstance) XML_GetUserData(parser);
	REALcontrolInstance instance = (REALcontrolInstance) ((void *)(parser));
	void (*fp)(REALcontrolInstance);
	fp = (void (*)(REALcontrolInstance)) REALGetEventInstance(instance, &xmlparserEvents[10]);
	if (fp) {
		fp (instance);
	}
}

static void metaEndCdataSection(XML_Parser parser)
{
	//REALcontrolInstance instance = (REALcontrolInstance) XML_GetUserData(parser);
	REALcontrolInstance instance = (REALcontrolInstance) ((void *)(parser));
	void (*fp)(REALcontrolInstance);
	fp = (void (*)(REALcontrolInstance)) REALGetEventInstance(instance, &xmlparserEvents[11]);
	if (fp) {
		fp (instance);
	}
}

static void beginandend(REALcontrolInstance instance, Boolean begin)
{
    int i=8; ////8=begin, 9=end
    if (!begin) i++;
   	void (*fp)(REALcontrolInstance);
	fp = (void (*)(REALcontrolInstance)) REALGetEventInstance(instance, &xmlparserEvents[i]);
	if (fp) {
		fp (instance);
	}
}

static void setHandlers (XML_Parser parser, REALobject instance)
{
	XML_SetUserData(parser, instance);
	XML_SetElementHandler(parser, startElement, endElement);
	XML_SetCharacterDataHandler(parser, characterData);
	XML_SetCommentHandler(parser, metaComment);
	XML_SetUnknownEncodingHandler(parser, unknownEncoding, 0);
	XML_SetDefaultHandlerExpand(parser, markup);	// !TT was: XML_SetDefaultHandler(...
	XML_SetProcessingInstructionHandler(parser, processingInstruction);
	XML_SetCdataSectionHandler(parser, metaStartCdataSection, metaEndCdataSection);
	XML_SetParamEntityParsing (parser, XML_PARAM_ENTITY_PARSING_ALWAYS);
	XML_SetUnknownEntityRefHandler (parser, unknownEntity);
}

/* ////THE CONSTRUCTOR FOR THE PARSER CLASS://////////

	This is where the parser is created, and all the data handlers are instantiated.

*/

static Boolean myreset2 (REALobject instance, REALstring enc)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (data->parser) {
		XML_ParserFree(data->parser);
		data->parser = nil;
	}
	data->amparsing = false;
	if (enc and enc->Length() > 0) {
		data->parser = XML_ParserCreate(enc->CString());
	} else {
		data->parser = XML_ParserCreate((const char*)NULL);
	}
	if (!(data->parser)) {
		/* this does, unfortunately, not propagate an exception in RB,
		 * as desired, but simply quit the app instead:
			REALobject exc = REALnewInstance ("NilObjectException");
			REALRaiseException (exc);
		 */
		return false;
	}
	data->depth = 0;
	setHandlers (data->parser, instance);
	return true;
}
	
static Boolean myreset (REALobject instance)
{
	return myreset2 (instance, nil);
}

static void xmlparserclassConstructor2 (REALobject instance, REALstring enc)
{
	myreset2 (instance, enc);
}

static void xmlparserclassConstructor (REALobject instance)
{
	xmlparserclassConstructor2 (instance, nil);
}

static void xmlparserclassDestructor (REALobject instance)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (data->parser) {
		XML_ParserFree(data->parser);
		data->parser = nil;
	}
	data->amparsing=false;
}

static long myparse (REALobject instance, REALstring s, Boolean done)
{
	if (s->Length() <= 0) return false;
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) return false;
	data->amparsing = true;	// !TT 29Apr02 moved here from the line behind "beginandend(...,true)"
	beginandend((REALcontrolInstance)instance,true);
	REALLockString(s);
	int success = XML_Parse(data->parser, REALCString(s), s->Length(), done);
	REALUnlockString(s);
	if (done) {
		beginandend((REALcontrolInstance)instance,false);
		data->amparsing=false;	// !TT 29Apr02 added
	}
	return success == 1;
}

static void myencoding (REALobject instance, REALstring enc)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if ((data->parser) && (enc->Length() > 0)) {
		XML_SetEncoding(data->parser, enc->CString());
	}
}
/* Possible encodings:
	ISO-8859-1
	US-ASCII
	UTF-8
	UTF-16

	// This is equivalent to supplying an encoding argument
	//to XML_CreateParser. It must not be called after XML_Parse
	//or XML_ParseBuffer. 

	int XMLPARSEAPI
	XML_SetEncoding(XML_Parser parser, const XML_Char *encoding);
*/
	

static REALstring myerrorstring (REALobject instance, int code)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) {
		return nil;
	}

	static const XML_LChar *message[] = {
		0,
		XML_T("out of memory"),
		XML_T("syntax error"),
		XML_T("no element found"),
		XML_T("not well-formed"),
		XML_T("unclosed token"),
		XML_T("unclosed token"),
		XML_T("mismatched tag"),
		XML_T("duplicate attribute"),
		XML_T("junk after document element"),
		XML_T("illegal parameter entity reference"),
		XML_T("undefined entity"),
		XML_T("recursive entity reference"),
		XML_T("asynchronous entity"),
		XML_T("reference to invalid character number"),
		XML_T("reference to binary entity"),
		XML_T("reference to external entity in attribute"),
		XML_T("xml processing instruction not at start of external entity"),
		XML_T("unknown encoding"),
		XML_T("encoding specified in XML declaration is incorrect"),
		XML_T("unclosed CDATA section"),
		XML_T("error in processing external entity reference"),
		XML_T("document is not standalone")
	};
	if (code > 0 && code < sizeof(message)/sizeof(message[0])) {
		REALstring str = REALBuildString(message[code],strlen(message[code]));
		REALSetStringEncoding (str, encUTF8);
		return str;
	}
	return nil;
}

static long errorGetter (REALobject instance, long param)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) {
		return -1000;
	}
	return XML_GetErrorCode(data->parser);
}

static long linenumberGetter (REALobject instance, long param)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) {
		return 0;
	}
	return XML_GetCurrentLineNumber(data->parser);
}
static long columnnumberGetter (REALobject instance, long param)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) {
		return 0;
	}
	return XML_GetCurrentColumnNumber(data->parser);
}
static long byteindexGetter (REALobject instance, long param)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) {
		return 0;
	}
	return XML_GetCurrentByteIndex(data->parser);
}
static long bytecountGetter (REALobject instance, long param)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (!data->parser) {
		return 0;
	}
	return XML_GetCurrentByteCount(data->parser);
}
	
static REALstring mybaseGetter (REALobject instance, long param)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if (data->parser) {
	  	const XML_Char *uri = XML_GetBase(data->parser);
	  	if (uri) {
			REALstring str = REALBuildString(uri,strlen(uri));
			REALSetStringEncoding (str, encUTF8);	// is this correct?
			return str;
		}
	}
	return nil;
}
static void mybaseSetter (REALobject instance, long param, REALstring uri)
{
	ClassData(xmlparserclass, instance, xmlparserclassData, data);
	if ((data->parser) && (uri->Length() > 0)) {
	  	//const XML_Char *uri = XML_GetBase(data->parser);
	  	XML_SetBase(data->parser, uri->CString());		// what about the encoding? shall it be preserved?
	}
}

static REALproperty xmlparserclassProperties[] = {
	{ nil, "Depth", "Integer", 0, REALstandardGetter, REALstandardSetter, FieldOffset(xmlparserclassData, depth)},
	{ nil, "Parsing", "Boolean", 0, REALstandardGetter, REALstandardSetter, FieldOffset(xmlparserclassData, amparsing)},
	{ nil, "ErrorCode", "Integer", REALpropRuntimeOnly, (REALproc) errorGetter, nil, 0 },
	{ nil, "LineNumber", "Integer", 0, (REALproc) linenumberGetter, nil, 0 },
	{ nil, "ColumnNumber", "Integer", 0, (REALproc) columnnumberGetter, nil, 0 },
	{ nil, "ByteIndex", "Integer", 0, (REALproc) byteindexGetter, nil, 0 },
	{ nil, "ByteCount", "Integer", 0, (REALproc) bytecountGetter, nil, 0 },
	{ nil, "Base", "String", 0, (REALproc) mybaseGetter, (REALproc) mybaseSetter, 0},
};

static REALmethodDefinition xmlparserclassMethods[] = {
	{ (REALproc) xmlparserclassConstructor, REALnoImplementation, ParserPluginName"()"},
	{ (REALproc) xmlparserclassConstructor2, REALnoImplementation, ParserPluginName"(encoding as String)"},
	{ (REALproc) myparse, REALnoImplementation, "Parse(data as String, done as Boolean) as Boolean"},
	{ (REALproc) myerrorstring, REALnoImplementation, "ErrorString(code as Integer) as String"},
	{ (REALproc) myreset, REALnoImplementation, "Reset() as Boolean"},
	{ (REALproc) myreset2, REALnoImplementation, "Reset(encoding as String) as Boolean"},
	{ (REALproc) myencoding, REALnoImplementation, "SetInputEncoding(enc as String)"},
};

static REALclassDefinition xmlparserclass = {
	kCurrentREALControlVersion,
	ParserPluginName,
	nil,
	sizeof(xmlparserclassData),
	0,
	nil, (REALproc) xmlparserclassDestructor,
	xmlparserclassProperties,
	sizeof(xmlparserclassProperties) / sizeof(REALproperty),
	xmlparserclassMethods,
	sizeof(xmlparserclassMethods) / sizeof(REALmethodDefinition),
	xmlparserEvents,
	sizeof(xmlparserEvents) / sizeof(REALevent)
};


static REALstring attrName (REALobject instance, long index)
{
	ClassData (xmlattrclass, instance, xmlattrclassData, data);
	if (index <= 0 or index > data->count) return nil;
	const XML_Char *key = data->atts[((index-1)<<1)];
	REALstring str = REALBuildString (key, strlen(key));
	REALSetStringEncoding (str, encUTF8);
	return str;
}

static REALstring attrValByIdx (REALobject instance, long index)
{
	ClassData (xmlattrclass, instance, xmlattrclassData, data);
	if (index <= 0 or index > data->count) return nil;
	const XML_Char *val = data->atts[((index-1)<<1)+1];
	REALstring str = REALBuildString (val, strlen(val));
	REALSetStringEncoding (str, encUTF8);
	convertLFtoCR (str);
	return str;
}

static REALstring attrValByName2 (REALobject instance, REALstring keyStr, REALstring defaultStr)
// note: attribute names in XML are case sensitive.
{
	ClassData (xmlattrclass, instance, xmlattrclassData, data);
	const XML_Char *targ = keyStr->CString();
	for (int index = 0; index < data->count; ++index) {
		const XML_Char *k = data->atts[index<<1];
		if (strcmp(k, targ) == 0) {
			const XML_Char *val = data->atts[(index<<1)+1];
			REALstring str = REALBuildString (val, strlen(val));
			REALSetStringEncoding (str, encUTF8);
			convertLFtoCR (str);
			return str;
		}
	}
	if (defaultStr) {
		REALLockString (defaultStr);
		return defaultStr;
	}
	return nil;
}

static REALstring attrValByName (REALobject instance, REALstring keyStr)
// note: attribute names in XML are case sensitive.
{
	return attrValByName2 (instance, keyStr, nil);
}

static REALproperty xmlattrclassProperties[] = {
	{ nil, "count", "Integer", 0, REALstandardGetter, REALstandardSetter, FieldOffset(xmlattrclassData, count)},
};

static REALmethodDefinition xmlattrclassMethods[] = {
	{ (REALproc) attrName, REALnoImplementation, "name(indexStartingAt1 as Integer) as String"},
	{ (REALproc) attrValByIdx, REALnoImplementation, "value(indexStartingAt1 as Integer) as String"},
	{ (REALproc) attrValByName, REALnoImplementation, "value(name as String) as String"},
	{ (REALproc) attrValByName2, REALnoImplementation, "value(name as String, default as String) as String"},
};

static REALclassDefinition xmlattrclass = {
	kCurrentREALControlVersion,
	AttrPluginName,
	nil,
	sizeof(xmlattrclassData),
	0,
	(REALproc) nil, (REALproc) nil,
	xmlattrclassProperties,
	sizeof(xmlattrclassProperties) / sizeof(REALproperty),
	xmlattrclassMethods,
	sizeof(xmlattrclassMethods) / sizeof(REALmethodDefinition)
};


void PluginEntry (void)
{
	REALRegisterClass (&xmlattrclass);
	REALRegisterClass (&xmlparserclass);
}
