/*
 * Plugin for REALbasic 4.5 and later to calculate CRCs
 *
 * This code is freeware, you may use and extend it as you like.
 *
 * The accompanying project file can be built with CodeWarrior Pro 6
 *
 * It was finished Jan 18, 2000 by Thomas Tempelmann.
 * Updated Apr 18, 2000: Added "...Cont" methods
 * Updated Mar 15, 2003: Update from CW Pro 4 to 6, added Carbon support.
 *
 * More RB-related stuff can be found here: <http://www.tempel.org/rb/>
 *
 * Enjoy and contribute!
 */

#include <string.h>
#include "rb_plugin.h"
#include "crc.h"

//-----------------------
// global plugin methods
//-----------------------

static long crc8Ptr (Ptr p, long l)
{
	return crc8 (p, l);
}

static long crc8cPtr (Ptr p, long l, long prev)
{
	return crc8 (p, l, prev);
}

static long crc8Str (REALstring s)
{
	return crc8 (s->CString(), s->Length());
}

static long crc8cStr (REALstring s, long prev)
{
	return crc8 (s->CString(), s->Length(), prev);
}


static long crcCCITTPtr (Ptr p, long l)
{
	return crcCCITT (p, l);
}

static long crcCCITTcPtr (Ptr p, long l, long prev)
{
	return crcCCITT (p, l, prev);
}

static long crcCCITTStr (REALstring s)
{
	return crcCCITT (s->CString(), s->Length());
}

static long crcCCITTcStr (REALstring s, long prev)
{
	return crcCCITT (s->CString(), s->Length(), prev);
}


static long crc32Ptr (Ptr p, long l)
{
	return crc32 (p, l);
}

static long crc32cPtr (Ptr p, long l, long prev)
{
	return crc32 (p, l, prev);
}

static long crc32Str (REALstring s)
{
	return crc32 (s->CString(), s->Length());
}

static long crc32cStr (REALstring s, long prev)
{
	return crc32 (s->CString(), s->Length(), prev);
}


static REALstring crcDillonPtr (long bits, Ptr p, long l)
{
	unsigned long long crc = crcDillon (bits, p, l);
	return REALBuildString ((Ptr)&crc, sizeof(crc));
}

static REALstring crcDillonStr (long bits, REALstring s)
{
	unsigned long long crc = crcDillon (bits, s->CString(), s->Length());
	return REALBuildString ((Ptr)&crc, sizeof(crc));
}


static REALmethodDefinition pluginMethodArray[] = {
	{ (REALproc) crc8Ptr,	REALnoImplementation, "CRC_8InMem(address as Integer, length as Integer) as Integer" },
	{ (REALproc) crc8cPtr,	REALnoImplementation, "CRC_8InMemCont(address as Integer, length as Integer, prevCRC as Integer) as Integer" },
	{ (REALproc) crc8Str,	REALnoImplementation, "CRC_8OfStr(s as String) as Integer" },
	{ (REALproc) crc8cStr,	REALnoImplementation, "CRC_8OfStrCont(s as String, prevCRC as Integer) as Integer" },

	{ (REALproc) crcCCITTPtr,	REALnoImplementation, "CRC_CCITTInMem(address as Integer, length as Integer) as Integer" },
	{ (REALproc) crcCCITTcPtr,	REALnoImplementation, "CRC_CCITTInMemCont(address as Integer, length as Integer, prevCRC as Integer) as Integer" },
	{ (REALproc) crcCCITTStr,	REALnoImplementation, "CRC_CCITTOfStr(s as String) as Integer" },
	{ (REALproc) crcCCITTcStr,	REALnoImplementation, "CRC_CCITTOfStrCont(s as String, prevCRC as Integer) as Integer" },
	{ (REALproc) crc32Ptr,		REALnoImplementation, "CRC_32InMem(address as Integer, length as Integer) as Integer" },
	{ (REALproc) crc32cPtr,		REALnoImplementation, "CRC_32InMemCont(address as Integer, length as Integer, prevCRC as Integer) as Integer" },
	{ (REALproc) crc32Str,		REALnoImplementation, "CRC_32OfStr(s as String) as Integer" },
	{ (REALproc) crc32cStr,		REALnoImplementation, "CRC_32OfStrCont(s as String, prevCRC as Integer) as Integer" },
	{ (REALproc) crcDillonPtr,	REALnoImplementation, "CRC_DillonInMem(bitWidth as Integer, address as Integer, length as Integer) as String" },
	{ (REALproc) crcDillonStr,	REALnoImplementation, "CRC_DillonOfStr(bitWidth as Integer, s as String) as String" },
};


//------------------------------------------
// extension methods to "MemoryBlock" class
//------------------------------------------
static long crc8MemBlk (REALmemoryBlock mb, long ofs, long l)
{
	return crc8 ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
}

static long crc8cMemBlk (REALmemoryBlock mb, long ofs, long l, long prev)
{
	return crc8 ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l, prev);
}

static long crcCCITTMemBlk (REALmemoryBlock mb, long ofs, long l)
{
	return crcCCITT ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
}

static long crcCCITTcMemBlk (REALmemoryBlock mb, long ofs, long l, long prev)
{
	return crcCCITT ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l, prev);
}

static long crc32MemBlk (REALmemoryBlock mb, long ofs, long l)
{
	return crc32 ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
}

static long crc32cMemBlk (REALmemoryBlock mb, long ofs, long l, long prev)
{
	return crc32 ((Ptr)REALMemoryBlockGetPtr(mb)+ofs, l, prev);
}

static REALstring crcDillonMemBlk (REALmemoryBlock mb, long bits, long ofs, long l)
{
	unsigned long long crc = crcDillon (bits, (Ptr)REALMemoryBlockGetPtr(mb)+ofs, l);
	return REALBuildString ((Ptr)&crc, sizeof(crc));
}

static REALmethodDefinition memBlkClassMethods[] = {
	{ (REALproc) crc8MemBlk,		REALnoImplementation, "CRC_8(offset as Integer, numBytes as Integer) as Integer"},
	{ (REALproc) crc8cMemBlk,		REALnoImplementation, "CRC_8Cont(offset as Integer, numBytes as Integer, prevCRC as Integer) as Integer"},
	{ (REALproc) crcCCITTMemBlk,	REALnoImplementation, "CRC_CCITT(offset as Integer, numBytes as Integer) as Integer"},
	{ (REALproc) crcCCITTcMemBlk,	REALnoImplementation, "CRC_CCITTCont(offset as Integer, numBytes as Integer, prevCRC as Integer) as Integer"},
	{ (REALproc) crc32MemBlk,		REALnoImplementation, "CRC_32(offset as Integer, numBytes as Integer) as Integer"},
	{ (REALproc) crc32cMemBlk,		REALnoImplementation, "CRC_32Cont(offset as Integer, numBytes as Integer, prevCRC as Integer) as Integer"},
	{ (REALproc) crcDillonMemBlk,	REALnoImplementation, "CRC_Dillon(bitWidth as Integer, offset as Integer, numBytes as Integer) as String"},
};

static REALclassDefinition memBlkClass = {
	kCurrentREALControlVersion,
	"MemoryBlock", nil, 0, 0, (REALproc) nil, (REALproc) nil, nil, 0,
	memBlkClassMethods, sizeof(memBlkClassMethods) / sizeof(REALmethodDefinition),
};


//-----------------------
// plugin initialization
//-----------------------

void PluginEntry (void)
{
	for (int i = 0; i < sizeof(pluginMethodArray) / sizeof(REALmethodDefinition); ++i) {
		REALRegisterMethod (&pluginMethodArray[i]);
	}
	
	REALRegisterClassExtension (&memBlkClass);
}
