//
// decoders.c
//
// contains the code that detects the encoding of the string
// and then calls the appropriate decoder
//
// by Thomas Tempelmann (rb@tempel.org), Mar 20, 2003


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "decoders.h"

static byte *text_ptr, *save_ptr;
static long iolen, save_len;

static void savePtr ()
{
	save_ptr = text_ptr;
	save_len = iolen;
}

static void restorePtr ()
{
	text_ptr = save_ptr;
	iolen = save_len;
}

static inline int atEOF ()
{
	return iolen <= 0;
}

static inline int inchar2 (void)
{
	if (iolen <= 0) {
		return EOF;
	}
	--iolen;
	return *text_ptr++;
}

int inchar (void)
{
	return inchar2();
}

static inline int nxtchar (void)
{
	if (iolen <= 0) {
		return EOF;
	}
	return *text_ptr;
}

// for some of the used algorithms we need to 0-terminate the lines
// but we must not do this persistently, since we'd alter the original string!
// so, we must preserve the change and restore it again before leaving here

byte* last_delim = NULL; byte last_ch;

static inline void restore_buf()
{
	if (last_delim) {
		*last_delim = last_ch;
		last_delim = NULL;
	}
}

int getline (byte* *line_start, long *line_len)
{
	restore_buf();

	if (atEOF()) return false;
	
	long len = 0;
	*line_start = text_ptr;
	do {
		byte ch = inchar2 ();
		if (ch == 13 or ch == 10) {
			if (ch == 13 and nxtchar() == 10) {
				inchar2 ();
			}
			break;
		}
		++len;
	} while (not atEOF());
	*line_len = len;
	return true;
}


byte* ad_fgetscr()
// 0-terminates the line
{
	byte *cp; long len;
	if (not getline (&cp, &len)) return NULL;

	last_delim = &cp[len];
	last_ch = *last_delim;
	*last_delim = 0; // marks the end of a line
	
	return cp;
}

static int allocBuf (void* *outbuf, long *outsize)
{
	*outsize = iolen * 4 / 3 + 100;
	*outbuf = malloc (*outsize);
	return *outbuf != NULL;
}

long smartDecode (byte* buf_start, long buf_len, char *filename, void* *outbuf_malloc)
{
	if (outbuf_malloc) *outbuf_malloc = NULL;
	text_ptr = buf_start;
	iolen = buf_len;
	*filename = 0;
	void* outbuf = NULL;
	long outsize;
	char* line;
	
	while ((line = (char*)ad_fgetscr()) != NULL) {
		if (*line == 0) {
			// skip empty lines
		} else if (strncmp(line,"=ybegin ",8) == 0) {
			//
			// yEnc encoded
			//

			char *cp=strstr(line,"name=");
			if (cp) {
				strcpy(filename,cp+5);  // Store the filename
				*cp=0; // throw away the filename
			}

			cp=strstr(line,"size=");
			if (cp==NULL) {
				//eprint("'size=' not found in =ybegin line. (%s)\r\n",line);
				restore_buf();
				return -1;  // Error - size not found
			}
			int y_size=atol(cp+5);

			cp=strstr(line,"line=");
			if (cp==NULL) {
				//eprint("'line=' not found in =ybegin line. (%s)\r\n",line);
				restore_buf();
				return -1;  // Error - linelength not found
			}
			int y_line=atol(cp+5);

			int y_part=0, y_parts=0;
			cp=strstr(line,"part=");   // Check if this is a multipart message
			if (cp) {
				y_part=atol(cp+5);
			}
			cp=strstr(line,"total=");
			if (cp) {
				y_parts=atol(cp+5);
			}
			
			if (!outbuf_malloc) {

				restore_buf();
				return 1;
			}
			
			if (not allocBuf (&outbuf, &outsize)) {
				restore_buf();
				return -25; // out of memory
			}
			
			int res = yDecode (y_line, y_size, y_part, y_parts, (byte*)outbuf, outsize);
			if (res > 0) res = -22; // something is wrong in the function, should not return values > 0
			if (res == 0) {
				res = y_size;
				*outbuf_malloc = outbuf;
			} else {
				free (outbuf);
			}
			restore_buf();
			return res;

		} else if (strncmp((char*)line, "begin ", 6) == 0) {
			//
			// uuencoded
			//

			int mode;
			sscanf ((char*)line, "begin %o %[^:]", &mode, filename);

			if (!outbuf_malloc) {
				restore_buf();
				return 2;
			}

			if (not allocBuf (&outbuf, &outsize)) return -25; // out of memory
			int res = uudecode ((byte*)outbuf, outsize);
			if (res > 0) {
				*outbuf_malloc = outbuf;
			} else {
				free (outbuf);
			}
			restore_buf();
			return res;

		} else if (strstr(line,"base64")) {
			//
			// appears to be base64 - base64 has no header, so we just skip after the next empty line and start there
			//
			
			// skip to empty line
			byte* line; long len;
			byte* cp;
			while (getline(&line, &len) != NULL and len > 0) {
				if ((cp = (byte*)strstr((char*)line,"filename=")) != NULL) {
					cp += 9;
					if (*cp == '"') {
						++cp;
						while (*cp >= ' ' and *cp != '"') *filename++ = *cp++;
					} else {
						while (*cp > ' ') *filename++ = *cp++;
					}
					*filename = 0;
				}
			}
			savePtr();
			while (getline(&line, &len) and len == 0) savePtr();
			restorePtr();
			
			if (!outbuf_malloc) {
				restore_buf();
				return 3;
			}
			if (not allocBuf (&outbuf, &outsize)) {
				restore_buf();
				return -25; // out of memory
			}
			int res = decodeBase64 ((byte*)outbuf, outsize);
			if (res > 0) {
				*outbuf_malloc = outbuf;
			} else {
				free (outbuf);
			}
			restore_buf();
			return res;
		}
	}
	restore_buf();
	return 0; // nothing found to decode
}
