/*
 * "SmartContentDecoder" plugin.
 *
 * Provide a method that can take text in the following formats and extract the binary file from it:
 *  - base64
 *  - uuencode
 *  - yEncode
 *
 * Written Mar 19, 2003 by Thomas Tempelmann (rb@tempel.org) for Ed Parsons (ed@tigersnake.com)
 */


#include <math.h>
#include <string.h>

#if TARGET_OS_WIN32
	static void inline DebugStr(unsigned char* p) {}
	#define log(x) OutputDebugString(x)
	#define PrepareCallback() 
	#define EnterCallback()
	#define ExitCallback()
#else
	#include <MacMemory.h>
	#include "dcon.h"
	#define log(x) dprintf(x)
#endif

#include "decoders.h"

#include "TT's Plugin Support.h"
#include "rb_plugin.h"


//-----------------------
// Some type definitions
//-----------------------

typedef long RBinteger;
typedef Boolean	RBboolVar;	// use this for Boolean variables stored in this plugin
typedef Boolean	RBboolPar;	// use this to receive Boolean parameters (both by value and by reference)
typedef long	RBboolRet;	// use this when returning a Boolean as a function result


//-----------------------
// Create a new RB class
//-----------------------


extern REALclassDefinition decoderClass;	// forward declaration

struct decoderClassData
{
	char filename[256];
	long errCode;
};

static void decoderClassConstructor (REALobject instance)
{
	ClassData (decoderClass, instance, decoderClassData, me);
}

static void decoderClassDestructor (REALobject instance)
{
	ClassData (decoderClass, instance, decoderClassData, me);
}


/* in case we want Event callbacks:

typedef RBinteger (*DecoderEventProc) (REALobject theObjInst, RBinteger someValue);	// this is the C version of the "DecoderEvent" declaration below, but with an added obj reference parm

static REALevent decoderClassEvents[] = {
	{ "DecoderEvent(someValue as Integer) as Integer" },
};

static RBinteger callDecoderEvent (REALobject instance, RBinteger firstArgument)
{
	RBinteger result;
	int idx = 0;			// this index selects from the decoderClassEvents array. 0 is the first entry.
	DecoderEventProc fp = (DecoderEventProc) REALGetEventInstance ((REALcontrolInstance)instance, &decoderClassEvents[idx]);
	if (fp) {
		result = fp (instance, firstArgument);
	} else {
		result = 0;
	}
	return result;
}
*/

static void RaiseOutOfMemoryException ()
{
	REALobject exc = REALnewInstance ("OutOfMemoryException");
	REALRaiseException (exc);
}

static REALstring decode (REALobject instance, REALstring in_str)
{
	ClassData (decoderClass, instance, decoderClassData, me);
	
	byte* buf_start = (byte*)in_str->CString();
	long buf_len = in_str->Length();
	
	void *outbuf;
	
	long result = smartDecode (buf_start, buf_len, me->filename, &outbuf);
	
	REALstring out_str = NULL;
	
	if (result > 0) {
		out_str = REALBuildString ((char*)outbuf, result);
		me->errCode = 0;
	} else {
		me->errCode = result;
	}
	
	if (outbuf) {
		free (outbuf);
	}

	return out_str;
}

static long guess (REALobject instance, REALstring in_str)
{
	ClassData (decoderClass, instance, decoderClassData, me);
	
	byte* buf_start = (byte*)in_str->CString();
	long buf_len = in_str->Length();
	
	long result = smartDecode (buf_start, buf_len, me->filename, NULL);

	if (result > 0) {
		me->errCode = 0;
	} else {
		result = 0;
		me->errCode = result;
	}

	return result;
}

static long error (REALobject instance)
{
	ClassData (decoderClass, instance, decoderClassData, me);

	return me->errCode;
}

static REALstring getfn (REALobject instance)
{
	ClassData (decoderClass, instance, decoderClassData, me);

	if (*me->filename) {
		return REALBuildString (me->filename, strlen(me->filename));
	}
	return NULL;
}

static REALmethodDefinition decoderClassMethods[] = {
	{ (REALproc) decode, REALnoImplementation, "Decode(text as String) as String"},
	{ (REALproc) guess, REALnoImplementation, "GuessEncoding(text as String) as Integer"},
	{ (REALproc) error, REALnoImplementation, "LastError as Integer"},
	{ (REALproc) getfn, REALnoImplementation, "LastFileName as String"},
};


static REALclassDefinition decoderClass = {
	kCurrentREALControlVersion,
	"SmartContentDecoder", nil,
	sizeof(decoderClassData), 0,
	(REALproc) decoderClassConstructor, (REALproc) decoderClassDestructor,
	nil, 0,
	decoderClassMethods, sizeof(decoderClassMethods) / sizeof(REALmethodDefinition),
	//decoderClassEvents, sizeof(decoderClassEvents) / sizeof(REALevent),
	nil, 0
};


// ---------------------------------
// ----- Plugin Initialization -----
// ---------------------------------

void PluginExit (void)
{
}

void PluginEntry (void)
{
	InitTTsPluginSupport ();

	REALRegisterClass (&decoderClass);
}
